from io import StringIO
from unittest.mock import patch
from unittest import TestCase
import os

import cat_win.cat as cat
import cat_win.util.ArgParser as ArgParser

# import sys
# sys.path.append('../cat_win')

test_file_dir = os.path.join(os.path.dirname(__file__), 'texts')
test_file_path  = os.path.join(test_file_dir, 'test.txt')
test_empty_path = os.path.join(test_file_dir, 'test_empty.txt')
test_peek       = os.path.join(test_file_dir, 'test_peek.txt')
test_result_B   = os.path.join(test_file_dir, 'full_test_result_B.txt')
test_result_C   = os.path.join(test_file_dir, 'full_test_result_C.txt')
test_result_D   = os.path.join(test_file_dir, 'full_test_result_D.txt')


class StdOutMock(StringIO):
    def reconfigure(self, encoding = None) -> None:
        return

    def fileno(self) -> int:
        return 0
    
# class StdInMock(DontReadFromInput):
class StdInMock:
    def reconfigure(self, encoding = None) -> None:
        return
    
    def readline(self) -> str:
        return ''


@patch('cat_win.cat.sys.stdin', StdInMock())
class TestCatFull(TestCase):
    maxDiff = None

    def tearDown(self):
        cat._CalculateLinePrefixSpacing.cache_clear()
        cat._CalculateLineLengthPrefixSpacing.cache_clear()
        ArgParser.FILE_ENCODING = 'utf-8'
        ArgParser.FILE_SEARCH = []
        ArgParser.FILE_MATCH = []
        ArgParser.FILE_TRUNCATE = [None, None, None]
        for i in range(len(cat.holder.args_id)):
            cat.holder.args_id[i] = False
    
    # no files parsed
    @patch('cat_win.cat.sys.argv', ['<CAT>', '-ln', '--nc', '[::-2]', 'enc=utf8'])
    def test_cat_output_full_A(self):
        expected_output = ''
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, '-ln', '--nc', '[::-2]', 'enc=utf8'])
    def test_cat_output_full_B(self):
        expected_output = ''
        with open(test_result_B, 'r', encoding='utf-8') as output:
            expected_output = output.read()
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', 'enc:UTF-8', test_file_path, '-ub', '[Sample,TEST]', '--nc', '-le'])
    def test_cat_output_full_C(self):
        expected_output = ''
        with open(test_result_C, 'r', encoding='utf-8') as output:
            expected_output = output.read()
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', 'enc=utf-8', test_file_path, 'trunc=2:6', '-n', '--reverse', '--nc', '-t'])
    def test_cat_output_full_D(self):
        expected_output = ''
        with open(test_result_D, 'r', encoding='utf-8') as output:
            expected_output = output.read()
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'trunc=0:0',])
    def test_cat_output_full_empty(self):
        expected_output = ''
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=cp1252', '--nc'])
    def test_cat_output_full_Cp1252(self):
        expected_output = ''
        with open(test_file_path, 'r', encoding='cp1252') as output:
            expected_output = output.read() + '\n'
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=latin_1', '--nc'])
    def test_cat_output_full_Latin1(self):
        expected_output = ''
        with open(test_file_path, 'r', encoding='latin_1') as output:
            expected_output = output.read() + '\n'
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=utf-8', 'trunc=0:1', 'find=ple ', 'match=:', '--nc'])
    def test_cat_output_full_find_found(self):
        expected_output = "Sample Text:\n--------------- Found [('ple ', [3, 7])] ---------------\n--------------- Matched [(':', [11, 12])] ---------------\n"
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=utf-8', 'trunc=0:1', 'find=NOTINLINE', '--nc'])
    def test_cat_output_full_find_not_found(self):
        expected_output = 'Sample Text:\n'
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)

    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=utf-8', 'trunc=0:1', 'find=Text', '--nc', '--nk'])
    def test_cat_output_full_find_found_nokeyword(self):
        expected_output = ''
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_file_path, 'enc=utf-8', 'find=Text', '--nc', '--grep'])
    def test_cat_output_full_find_found_grep(self):
        expected_output = 'Sample Text:\n'
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            
    @patch('cat_win.cat.sys.argv', ['<CAT>', test_peek, 'enc=utf-8', '--peek'])
    def test_cat_output_full_peek(self):
        expected_output = """1
2
3
4
5
           •
          (11)
           •
6
7
8
9
10
"""
        with patch('sys.stdout', new=StdOutMock()) as fake_out:
            cat.main()
            self.assertEqual(fake_out.getvalue(), expected_output)
            

# python -m unittest discover -s tests -p test*.py


try:
    from colorama import init as coloramaInit
except ImportError:
    nop = lambda *_, **__: None; coloramaInit = nop
from datetime import datetime
from functools import lru_cache
from itertools import groupby
from platform import system
from re import sub as resub
import os
import sys

from cat_win.const.ArgConstants import *
from cat_win.const.ColorConstants import C_KW
from cat_win.util.Base64 import decodeBase64, encodeBase64
from cat_win.util.FileAttributes import getFileMetaData, getFileSize, _convert_size
from cat_win.util.RawViewer import getRawViewLinesGen
from cat_win.web.UpdateChecker import printUpdateInformation
import cat_win.persistence.Config as Config
import cat_win.util.ArgParser as ArgParser
import cat_win.util.checksum as checksum
import cat_win.util.Converter as Converter
import cat_win.util.Holder as Holder
import cat_win.util.StdInHelper as StdInHelper
import cat_win.util.StringFinder as StringFinder
import cat_win.util.TmpFileHelper as TmpFileHelper

from cat_win import __project__, __version__, __sysversion__, __author__, __url__
workingDir = os.path.dirname(os.path.realpath(__file__))

coloramaInit()
config = Config.Config(workingDir)

color_dic = config.loadConfig()

converter = Converter.Converter()
holder = Holder.Holder()
tmpFileHelper = TmpFileHelper.TmpFileHelper()

on_windows_os = (system() == 'Windows')

def exception_handler(exception_type: type, exception, traceback, debug_hook=sys.excepthook) -> None:
    print(color_dic[C_KW.RESET_ALL])
    if holder.args_id[ARGS_DEBUG]:
        debug_hook(exception_type, exception, traceback)
        return
    print(f"\n{exception_type.__name__}{':' * bool(str(exception))} {exception}")
    if exception_type != KeyboardInterrupt:
        print(f"If this Exception is unexpected, please raise an official Issue at:\n{__url__}/issues")


sys.excepthook = exception_handler


def _showHelp() -> None:
    """
    Show the Help message and exit.
    """
    helpMessage = 'Usage: catw [FILE]... [OPTION]...\n'
    helpMessage += 'Concatenate FILE(s) to standard output.\n\n'
    for x in ALL_ARGS:
        helpMessage += f"\t{f'{x.shortForm}, {x.longForm}': <25}{x.help}\n"
    helpMessage += '\n'
    helpMessage += f"\t{'enc=X, enc:X'    : <25}set file encoding to X (default is utf-8)\n"
    helpMessage += f"\t{'find=X, find:X'  : <25}find/query a substring X in the given files\n"
    helpMessage += f"\t{'match=X, match:X': <25}find/query a pattern X in the given files\n"
    helpMessage += f"\t{'trunc=X:Y, trunc:X:Y': <25}truncate file to lines x and y (python-like)\n"
    helpMessage += '\n'
    helpMessage += f"\t{'[a,b]': <25}replace a with b in every line\n"
    helpMessage += f"\t{'[a:b:c]': <25}python-like string indexing syntax (line by line)\n"
    helpMessage += '\n'
    helpMessage += 'Examples:\n'
    helpMessage += f"\t{'cat f g -r' : <25}Output g's contents in reverse order, then f's content in reverse order\n"
    helpMessage += f"\t{'cat f g -ne': <25}Output f's, then g's content, while numerating and showing the end of lines\n"
    helpMessage += f"\t{'cat f trunc=a:b:c': <25}Output f's content starting at line a, ending at line b, stepping c\n"
    print(helpMessage)
    printUpdateInformation(__project__, __version__, color_dic)


def _showVersion() -> None:
    """
    Show the Version message and exit.
    """
    catVersion = f"Catw {__version__} - from {workingDir}\n"
    versionMessage = '\n'
    versionMessage += '-' * len(catVersion) + '\n'
    versionMessage += catVersion
    versionMessage += '-' * len(catVersion) + '\n'
    versionMessage += '\n'
    versionMessage += f"Built with: \tPython {__sysversion__}\n"  # sys.version
    try:
        versionMessage += f"Install time: \t{datetime.fromtimestamp(os.path.getctime(os.path.realpath(__file__)))}\n"
    except OSError: # fails on pyinstaller executable
        versionMessage += f"Install time: \t-\n"
    versionMessage += f"Author: \t{__author__}\n"
    print(versionMessage)
    printUpdateInformation(__project__, __version__, color_dic)


def _showDebug(args: list, unknown_args: list, known_files: list, unknown_files: list, echo_args: list) -> None:
    """
    Print all neccassary debug information
    """
    print('Debug Information:')
    print('args: ', end='')
    print([(arg[0], arg[1], holder.args_id[arg[0]]) for arg in args])
    print('unknown_args: ', end='')
    print(unknown_args)
    print('known_files: ', end='')
    print(known_files)
    print('unknown_files: ', end='')
    print(unknown_files)
    print('echo_args: ', end='')
    print(echo_args)
    print('file encoding: ', end='')
    print(ArgParser.FILE_ENCODING)
    print('search keyword(s): ', end='')
    print(ArgParser.FILE_SEARCH)
    print('search match(es): ', end='')
    print(ArgParser.FILE_MATCH)
    print('truncate file: ', end='')
    print(ArgParser.FILE_TRUNCATE)


def _showCount() -> None:
    """
    display the line sum of each file individually if
    ARGS_CCOUNT is specified.
    display the line sum of all files
    """
    if holder.args_id[ARGS_CCOUNT]:
        longestFileName = max(map(len, holder.allFilesLines.keys())) + 1
        print(f"{color_dic[C_KW.COUNT_AND_FILES]}{'File': <{longestFileName}}LineCount{color_dic[C_KW.RESET_ALL]}")
        for file in holder.allFilesLines.keys():
            print(f"{color_dic[C_KW.COUNT_AND_FILES]}{file: <{longestFileName}}{holder.allFilesLines[file]: >{holder.fileLineNumberPlaceHolder}}{color_dic[C_KW.RESET_ALL]}")
        print('')
    print(f"{color_dic[C_KW.COUNT_AND_FILES]}Lines (Sum): {holder.allFilesLinesSum}{color_dic[C_KW.RESET_ALL]}")


def _showFiles() -> None:
    """
    displays holder.files including their size and calculates
    their size sum.
    """
    if len(holder.files) == 0:
        return
    file_sizes = []
    msg = 'found' if holder.args_id[ARGS_FFILES] else 'applied'
    print(color_dic[C_KW.COUNT_AND_FILES], end='')
    print(f"{msg} FILE(s):", end='')
    print(color_dic[C_KW.RESET_ALL])
    for file in holder.files:
        size = getFileSize(file.path)
        file_sizes.append(size)
        print(f"\t{color_dic[C_KW.COUNT_AND_FILES]}{_convert_size(size): <10}{'*' if file.containsQueried else ' '}{file.displayname}{color_dic[C_KW.RESET_ALL]}")
    print(color_dic[C_KW.COUNT_AND_FILES], end='')
    print(f"Sum:\t{_convert_size(sum(file_sizes))}", end='')
    print(color_dic[C_KW.RESET_ALL])
    print(color_dic[C_KW.COUNT_AND_FILES], end='')
    print(f"Amount:\t{len(holder.files)}", end='')
    print(color_dic[C_KW.RESET_ALL])


def _printMeta(file: str) -> None:
    """
    print the information retrieved by getFileMetaData()
    
    Parameters:
    file (str):
        a string representation of a file (-path)
    """
    metaData = getFileMetaData(file, on_windows_os,
                               [color_dic[C_KW.RESET_ALL],
                               color_dic[C_KW.ATTRIB],
                               color_dic[C_KW.ATTRIB_POSITIVE],
                               color_dic[C_KW.ATTRIB_NEGATIVE]])
    print(metaData)


def _printChecksum(file: str) -> None:
    """
    print the information retrieved by getChecksumFromFile()
    
    Parameters:
    file (str):
        a string representation of a file (-path)
    """
    print(f"{color_dic[C_KW.CHECKSUM]}Checksum of '{file}':{color_dic[C_KW.RESET_ALL]}")
    print(checksum.getChecksumFromFile(file, [color_dic[C_KW.CHECKSUM], color_dic[C_KW.RESET_ALL]]))


def _printMetaAndChecksum(showMeta: bool, showChecksum: bool) -> None:
    """
    calls _printMeta() and _printChecksum() on every file.
    
    Parameters:
    showMeta (bool):
        decides if the metadata of the files should be displayed
    showChecksum (bool):
        decides if the checksum of the files should be displayed
    """
    for file in holder.files:
        if showMeta:
            _printMeta(file.path)
        if showChecksum:
            _printChecksum(file.path)


def removeAnsiCodesFromLine(line: str) -> str:
    """
    Parameters:
    line (str):
        the string to clean ANSI-Colorcodes from
        
    Returns
    (str):
        the cleaned string
    """
    # version 1: efficiency is about the same, and does not have any dependency
    # however it is not as safe in case of unusual/broken escape sequences.
    # while (codePosStart := line.find(ESC_CODE)) != -1:
    #     codePosEnd = line[codePosStart:].find('m')
    #     # here should be checks like 'codePosEnd' != -1 and
    #     # 'codePosEnd' <= 5 to make sure we found a valid EscapeSequence
    #     # for a better performance let's assume all EscapeSequences are valid...
    #     line = line[:codePosStart] + line[codePosStart+codePosEnd+1:]
    # return line
    # version 2:
    return resub(r'\x1b\[[0-9\;]*m', '', line)


# def removeAnsiCodes(content: list) -> list:
#     return [(removeAnsiCodesFromLine(prefix), removeAnsiCodesFromLine(line)) for prefix, line in content]


@lru_cache(maxsize=None)
def _CalculateLinePrefixSpacing(lineCharLength: int,
                                includeFilePrefix: bool = False, fileCharLength: int = 0) -> str:
    """
    calculate a string template for the line prefix.
    
    Parameters:
    lineCharLength (int):
        the length of the line number
    includeFilePrefix (bool):
        should the file be included in the prefix
    fileCharLength (int):
        the length of the file number
    
    Returns:
    (str):
        a non-finished but correctly formatted string template to insert line number
        and file index into
    """
    line_prefix = (' ' * (holder.fileLineNumberPlaceHolder - lineCharLength)) + '%i)'

    if includeFilePrefix:
        file_prefix = (' ' * (holder.fileNumberPlaceHolder - fileCharLength)) + '%i.'
        return color_dic[C_KW.NUMBER] + file_prefix + line_prefix + color_dic[C_KW.RESET_ALL] + ' '

    return color_dic[C_KW.NUMBER] + line_prefix + color_dic[C_KW.RESET_ALL] + ' '


def _getLinePrefix(line_num: int, index: int) -> str:
    """
    calculates the line prefix in regard to the line number and file count.
    
    Parameters:
    line_num (int):
        the current number identifying the line
    index (int):
        the current number identifying the file
    
    Returns:
    (str):
        the new line prefix including the line number.
    """
    if len(holder.files) > 1:
        return _CalculateLinePrefixSpacing(len(str(line_num)), True, len(str(index))) % (index, line_num)
    return _CalculateLinePrefixSpacing(len(str(line_num))) % (line_num)


@lru_cache(maxsize=None)
def _CalculateLineLengthPrefixSpacing(lineCharLength: int) -> str:
    """
    calculate a string template for the line prefix.
    
    Parameters:
    lineCharLength (int):
        the length of the line
    
    Returns:
    (str):
        a non-finished but correctly formatted string template to insert line length into
    """
    lengthPrefix = '[' + ' ' * (holder.fileLineLengthPlaceHolder - lineCharLength) + '%i]'
    return '%s' + color_dic[C_KW.LINE_LENGTH] + lengthPrefix + color_dic[C_KW.RESET_ALL] + ' '


def _getLineLengthPrefix(prefix: str, line) -> str:
    """
    calculates the line prefix in regard to the line length.
    
    Parameters:
    prefix (str):
        the current prefix to append to
    line (str|byte):
        a representation of the current line
    
    Returns:
    (str):
        the new line prefix including the line length.
    """
    if not holder.args_id[ARGS_NOCOL] and type(line) == str:
        line = removeAnsiCodesFromLine(line)
    return _CalculateLineLengthPrefixSpacing(len(str(len(line)))) % (prefix, len(line))


def printFile(content: list) -> bool:
    """
    print a file and possibly include the substrings and patterns to search for.
    
    Parameters:
    content (list):
        the content of a file like [(prefix, line), ...]
        
    Returns
    (bool):
        identifies if the given content parameter contained any
        queried keyword/pattern.
    """
    if not content:
        return False
    if not (ArgParser.FILE_SEARCH or ArgParser.FILE_MATCH):
        print(*[prefix + line for prefix, line in content], sep='\n')
        return False

    containsQueried = False
    stringFinder = StringFinder.StringFinder(ArgParser.FILE_SEARCH, ArgParser.FILE_MATCH)

    for line_prefix, line in content:
        cleanedLine = removeAnsiCodesFromLine(line)
        intervals, fKeyWords, mKeywords = stringFinder.findKeywords(cleanedLine)
        
        if len(fKeyWords + mKeywords) == 0:
            if not holder.args_id[ARGS_GREP]:
                print(line_prefix + line)
            continue
        
        containsQueried = True
        if holder.args_id[ARGS_NOKEYWORD]:
            continue
        
        if not holder.args_id[ARGS_NOCOL]:
            for kw_pos, kw_code in intervals:
                cleanedLine = cleanedLine[:kw_pos] + color_dic[kw_code] + cleanedLine[kw_pos:]

        print(line_prefix + cleanedLine)

        if holder.args_id[ARGS_GREP] or holder.args_id[ARGS_NOBREAK]:
            continue
        
        found_sth = False
        if fKeyWords:
            print(color_dic[C_KW.FOUND_MESSAGE], end='')
            print('--------------- Found', fKeyWords, '---------------', end='')
            print(color_dic[C_KW.RESET_ALL])
            found_sth = True
        if mKeywords:
            print(color_dic[C_KW.MATCHED_MESSAGE], end='')
            print('--------------- Matched', mKeywords, '---------------', end='')
            print(color_dic[C_KW.RESET_ALL])
            found_sth = True

        if found_sth:
            try:  # fails when using -i mode, because the stdin will send en EOF char to input without prompting the user
                input()
            except (EOFError, UnicodeDecodeError):
                pass
            
    return containsQueried


def printExcludedByPeek(content: list, excludedByPeek: int) -> None:
    """
    print a paragraph about how many lines have been excluded.
    
    Parameters:
    content (list):
        the content of a file like [(prefix, line), ...]
    excludedByPeek (int):
        the amount of lines that have been excluded
    """
    if not excludedByPeek or len(content) <= 5:
        return
    excludedByPeek = excludedByPeek + 10 - len(content)
    prefix = content[0][0]
    prefix = prefix.replace(color_dic[C_KW.NUMBER], '')
    prefix = prefix.replace(color_dic[C_KW.LINE_LENGTH], '')
    prefix = prefix.replace(color_dic[C_KW.RESET_ALL], '')
    excludedByPeekLength = (len(str(excludedByPeek))-1)//2
    excludedByPeekIndent = ' ' * (len(prefix) - excludedByPeekLength + 10)
    excludedByPeekIndentAdd = ' ' * excludedByPeekLength
    print(color_dic[C_KW.NUMBER], end='')
    print(excludedByPeekIndent, excludedByPeekIndentAdd, ' •', sep='')
    print(excludedByPeekIndent, '(', excludedByPeek, ')', sep='')
    print(excludedByPeekIndent, excludedByPeekIndentAdd, ' •', sep='', end='')
    print(color_dic[C_KW.RESET_ALL])


def editContent(content: list, show_bytecode: bool, fileIndex: int = 0, lineOffset: int = 0) -> None:
    """
    apply all parameters to a string (file Content).
    
    Parameters:
    content (list):
        the content of a file like [(prefix, line), ...]
    show_bytecode (bool).
        indicates if the content lines are string or bytes
    fileIndex (int):
        the index of the holder.files list, pointing to the file that
        is currently being processed. a negative value can be used for
        the shell mode
    """
    excludedByPeek = 0
    
    if not show_bytecode and holder.args_id[ARGS_B64D]:
        content = decodeBase64(content, ArgParser.FILE_ENCODING)
    
    if holder.args_id[ARGS_NUMBER]:
        content = [(_getLinePrefix(j+lineOffset, fileIndex+1), c[1])
                   for j, c in enumerate(content, start=1)]
    content = content[ArgParser.FILE_TRUNCATE[0]:ArgParser.FILE_TRUNCATE[1]:ArgParser.FILE_TRUNCATE[2]]
    if holder.args_id[ARGS_PEEK] and len(content) > 10:
        excludedByPeek = len(content) - 10
        content = content[:5] + content[-5:]

    if not show_bytecode:
        for arg, param in holder.args:
            if arg == ARGS_CUT:
                try:
                    content = [(prefix, eval(repr(line) + param))
                                for prefix, line in content]
                except:
                    print('Error at operation: ', param)
                    return
        
        for arg, param in holder.args:
            if arg == ARGS_ENDS:
                content = [(prefix, line + color_dic[C_KW.ENDS] + '$' + color_dic[C_KW.RESET_ALL])
                           for prefix, line in content]
            elif arg == ARGS_TABS:
                content = [(prefix, line.replace('\t', color_dic[C_KW.TABS] + '^I' + color_dic[C_KW.RESET_ALL]))
                           for prefix, line in content]
            elif arg == ARGS_SQUEEZE:
                content = [list(group)[0] for _, group in groupby(content, lambda x: x[1])]
            elif arg == ARGS_REVERSE:
                content.reverse()
            elif arg == ARGS_BLANK:
                content = [c for c in content if c[1]]
            elif arg == ARGS_DEC:
                content = [(prefix, line + ' ' + color_dic[C_KW.CONVERSION] + converter._fromDEC(int(line), (param == param.lower())) +
                            color_dic[C_KW.RESET_ALL]) for prefix, line in content if converter.is_dec(line)]
            elif arg == ARGS_HEX:
                content = [(prefix, line + ' ' + color_dic[C_KW.CONVERSION] + converter._fromHEX(line, (param == param.lower())) +
                            color_dic[C_KW.RESET_ALL]) for prefix, line in content if converter.is_hex(line)]
            elif arg == ARGS_BIN:
                content = [(prefix, line + ' ' + color_dic[C_KW.CONVERSION] + converter._fromBIN(line, (param == param.lower())) +
                            color_dic[C_KW.RESET_ALL]) for prefix, line in content if converter.is_bin(line)]
            elif arg == ARGS_REPLACE:
                replace_values = param[1:-1].split(",")
                content = [(prefix, line.replace(replace_values[0], color_dic[C_KW.REPLACE] + replace_values[1] + color_dic[C_KW.RESET_ALL]))
                           for prefix, line  in content]
            elif arg == ARGS_EOF:
                content = [(prefix, line.replace(chr(26), color_dic[C_KW.REPLACE] + '^EOF' + color_dic[C_KW.RESET_ALL]))
                           for prefix, line in content]
            
    if holder.args_id[ARGS_LLENGTH]:
        content = [(_getLineLengthPrefix(c[0], c[1]), c[1]) for c in content]
    if holder.args_id[ARGS_B64E]:
        content = encodeBase64(content, ArgParser.FILE_ENCODING)

    foundQueried = printFile(content[:len(content)//2])
    if fileIndex >= 0:
        holder.files[fileIndex].setContainsQueried(foundQueried)
    printExcludedByPeek(content, excludedByPeek)
    foundQueried = printFile(content[len(content)//2:])
    if fileIndex >= 0:
        holder.files[fileIndex].setContainsQueried(foundQueried)
    
    if not show_bytecode:
        if holder.args_id[ARGS_CLIP]:
            holder.clipBoard += '\n'.join([prefix + line for prefix, line in content])


def editFile(fileIndex: int = 0) -> None:
    """
    apply all parameters to a file.
    
    Parameters:
    fileIndex (int):
        the index regarding which file is currently being edited
    """
    show_bytecode = False
    
    content = [('', '')]
    try:
        with open(holder.files[fileIndex].path, 'r', encoding=ArgParser.FILE_ENCODING) as f:
            # splitlines() gives a slight inaccuracy, in case the last line is empty.
            # the alternative would be worse: split('\n') would increase the linecount each
            # time catw touches a file.
            content = [('', line) for line in f.read().splitlines()]
    except:
        print('Failed to open:', holder.files[fileIndex].displayname)
        try:
            enterChar = '⏎'
            try:
                if len(enterChar.encode(ArgParser.FILE_ENCODING)) != 3:
                    raise UnicodeEncodeError('', '', -1, -1, '')
            except UnicodeEncodeError:
                enterChar = 'ENTER'
            print(f"Do you want to open the file as a binary, without parameters? [Y/{enterChar}]:", end='')
            inp = input()
            if not 'Y' in inp.upper() and inp:
                print('Aborting...')
                return
        except EOFError:
            pass
        except UnicodeError:
            print(f"Input is not recognized in the given encoding: {ArgParser.FILE_ENCODING}")
            print('Aborting...')
            return
        try:
            with open(holder.files[fileIndex].path, 'rb') as f:
                # in binary splitlines() is our only option
                content = [('', repr(line)[2:-1]) for line in f.read().splitlines()]
            show_bytecode = True
        except:
            print('Operation failed! Try using the enc=X parameter.')
            return
    
    editContent(content, show_bytecode, fileIndex)


def _copyToClipboard(content: str, __dependency: int = 3, __clipBoardError: bool = False) -> None:
    """
    copy a string to the clipboard, by recursively checking which module exists and could
    be used, this function should only be called by copyToClipboard()
    
    Parameters:
    content (str):
        the string to copy
    __dependency (int):
        do not change!
    __clipBoardError (bool):
        do not change!
    """
    if __dependency == 0:
        if __clipBoardError:
            errorMsg = '\n'
            errorMsg += "ClipBoardError: You can use either 'pyperclip3', 'pyperclip', or 'pyclip' in order to use the '--clip' parameter.\n"
            errorMsg += "Try to install a different one using 'python -m pip install ...'"
        else:
            errorMsg = '\n'
            errorMsg += "ImportError: You need either 'pyperclip3', 'pyperclip', or 'pyclip' in order to use the '--clip' parameter.\n"
            errorMsg += "Should you have any problem with either module, try to install a different one using 'python -m pip install ...'"
        print(errorMsg)
        return
    try:
        if __dependency == 3:
            import pyperclip as pc
        elif __dependency == 2:
            import pyclip as pc
        elif __dependency == 1:
            import pyperclip3 as pc
        pc.copy(content)
    except ImportError:
        _copyToClipboard(content, __dependency-1, False or __clipBoardError)
    except Exception:
        _copyToClipboard(content, __dependency-1, True or __clipBoardError)
    

def copyToClipboard(content: str) -> None:
    """
    entry point to recursive function _copyToClipboard()
    
    Parameters:
    content (str):
        the string to copy
    """
    _copyToClipboard(content)


def printRawView(fileIndex: int = 0, mode: str = 'X') -> None:
    """
    print the raw byte representation of a file in hexadecimal or binary
    
    Parameters:
    fileIndex (int):
        the index regarding which file is currently being edited
    mode (str):
        either 'x', 'X' for hexadecimal (lower- or upper case letters),
        or 'b' for binary
    """
    print(holder.files[fileIndex].displayname, ':', sep='')
    for line in getRawViewLinesGen(holder.files[fileIndex].path, mode, [color_dic[C_KW.RAWVIEWER], color_dic[C_KW.RESET_ALL]], ArgParser.FILE_ENCODING):
        print(line)
    print('')


def editFiles() -> None:
    """
    manage the calls to editFile() for each file.
    """
    start = len(holder.files)-1 if holder.reversed else 0
    end = -1 if holder.reversed else len(holder.files)

    rawViewMode = None
    if holder.args_id[ARGS_HEXVIEW] or holder.args_id[ARGS_BINVIEW]:
        for arg, param in holder.args:
            if arg == ARGS_HEXVIEW:
                rawViewMode = 'X' if param == param.upper() else 'x'
                break
            elif arg == ARGS_BINVIEW:
                rawViewMode = 'b'
                break
    
    for i in range(start, end, -1 if holder.reversed else 1):
        if rawViewMode:
            printRawView(i, rawViewMode)
        else:
            editFile(i)
    if holder.args_id[ARGS_COUNT]:
        print('')
        _showCount()
    if holder.args_id[ARGS_FILES]:
        print('')
        _showFiles()
    if holder.args_id[ARGS_CLIP]:
        copyToClipboard(removeAnsiCodesFromLine(holder.clipBoard))


def init(shell: bool = False) -> tuple:
    """
    initiate the code by calling the argparser and handling the default
    parameters: -h, -v, -d, --config.
    
    Parameters:
    contenshellt (bool):
        indicates if the shell entry point was used, and the stdin will therefor
        be used by default
        
    Returns
    (tuple):
        contains (known_files, unknown_files, echo_args) from the argparser
    """
    # read parameter-args
    args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(sys.argv)

    holder.setArgs(args)

    if holder.args_id[ARGS_RECONFIGURE]:
        sys.stdout.reconfigure(encoding=ArgParser.FILE_ENCODING)
        sys.stdin.reconfigure(encoding=ArgParser.FILE_ENCODING)
    
    # do not use colors if requested, or output will be piped anyways
    if holder.args_id[ARGS_NOCOL] or (not sys.stdout.isatty() or sys.stdout.closed):
        global color_dic
        color_dic = dict.fromkeys(color_dic,'')
    
    # check for special cases
    if holder.args_id[ARGS_DEBUG]:
        _showDebug(args, unknown_args, known_files, unknown_files, echo_args)
    if (len(known_files) + len(unknown_files) + len(holder.args) == 0 and not shell) or holder.args_id[ARGS_HELP]:
        _showHelp()
        sys.exit(0)
    if holder.args_id[ARGS_VERSION]:
        _showVersion()
        sys.exit(0)
    if holder.args_id[ARGS_CONFIG]:
        config.saveConfig()
        sys.exit(0)
    
    return (known_files, unknown_files, echo_args)


def main():
    piped_input = temp_file = ''
    known_files, unknown_files, echo_args = init(False)
    
    if holder.args_id[ARGS_ECHO]:
        temp_file = StdInHelper.writeTemp(' '.join(echo_args), tmpFileHelper.generateTempFileName(), ArgParser.FILE_ENCODING)
        known_files.append(temp_file)
        holder.setTempFileEcho(temp_file)
    if holder.args_id[ARGS_INTERACTIVE]:
        piped_input = ''.join(StdInHelper.getStdInContent(holder.args_id[ARGS_ONELINE]))
        temp_file = StdInHelper.writeTemp(piped_input, tmpFileHelper.generateTempFileName(), ArgParser.FILE_ENCODING)
        known_files.append(temp_file)
        unknown_files = StdInHelper.writeFiles(unknown_files, piped_input, ArgParser.FILE_ENCODING)
        holder.setTempFileStdIn(temp_file)
    else:
        unknown_files = StdInHelper.readWriteFilesFromStdIn(
            unknown_files, ArgParser.FILE_ENCODING, on_windows_os, holder.args_id[ARGS_ONELINE])

    if (len(known_files) + len(unknown_files) == 0):
        return

    # fill holder object with neccessary values
    holder.setFiles([*known_files, *unknown_files])
    
    if holder.args_id[ARGS_FFILES]:
        _showFiles()
        return
    if holder.args_id[ARGS_DATA] or holder.args_id[ARGS_CHECKSUM]:
        _printMetaAndChecksum(holder.args_id[ARGS_DATA], holder.args_id[ARGS_CHECKSUM])
        return
    
    if holder.args_id[ARGS_B64D]:
        holder.setDecodingTempFiles([tmpFileHelper.generateTempFileName() for _ in holder.files])
    holder.generateValues(ArgParser.FILE_ENCODING)

    if holder.args_id[ARGS_CCOUNT]:
        _showCount()
        return

    try:
        editFiles()  # print the cat-output
    except IOError: # catch broken-pipe error
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE

    # clean-up
    for tmp_file in tmpFileHelper.getGeneratedTempFiles():
        if holder.args_id[ARGS_DEBUG]:
            print('Cleaning', tmp_file)
        try:
            os.remove(tmp_file)
        except FileNotFoundError:
            if holder.args_id[ARGS_DEBUG]:
                print('FileNotFoundError', tmp_file)


def shell_main():
    init(True)
    
    shellPrefix = '>>> '
    EOFControlChar = 'Z' if on_windows_os else 'D'
    
    print(__project__, 'v' + __version__, 'shell', '(' + __url__ + ')')
    print(f"Use 'catw' to handle files. Type ^{EOFControlChar} (Ctrl + {EOFControlChar}) to exit.")
    
    print(shellPrefix, end='', flush=True)
    for i, line in enumerate(StdInHelper.getStdInContent(holder.args_id[ARGS_ONELINE])):
        editContent([('', line.rstrip('\n'))], False, -1, i)
        if not holder.args_id[ARGS_ONELINE]:
            print(shellPrefix, end='', flush=True)


if __name__ == '__main__':
    main()



from unittest import TestCase
import os

from cat_win.const.ArgConstants import *
import cat_win.util.ArgParser as ArgParser
# import sys
# sys.path.append('../cat_win')


test_file_dir = os.path.join(os.path.dirname(__file__), 'texts')



class TestArgParser(TestCase):
    maxDiff = None
    
    def tearDown(self):
        ArgParser.FILE_ENCODING = 'utf-8'
        ArgParser.FILE_MATCH = []
        ArgParser.FILE_SEARCH = []
        ArgParser.FILE_TRUNCATE = [None, None, None]
        
    def test_getArguments_empty(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT'])
        self.assertCountEqual(args, [])
        self.assertCountEqual(unknown_args, [])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_duplicate(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', '-n', '-n', '-c'])
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, ['-n', '-c', '-n'])
        self.assertCountEqual(unknown_args, [])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_concatenated(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', '-abcdef'])
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, ['-a', '-b', '-c', '-d', '-e', '-f'])
        self.assertCountEqual(unknown_args, [])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_concatenated_unknown(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', '-+abcde?fϵg'])
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, ['-a', '-b', '-c', '-d', '-e', '-f', '-g'])
        self.assertCountEqual(unknown_args, ['-+', '-?', '-ϵ'])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_concatenated_invalid(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', '--abcde?fg'])
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, [])
        self.assertCountEqual(unknown_args, ['--abcde?fg'])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_encoding(self):
        ArgParser.getArguments(['CAT', 'enc:ascii'])
        self.assertEqual(ArgParser.FILE_ENCODING, 'ascii')
        ArgParser.getArguments(['CAT', 'enc=utf-8'])
        self.assertEqual(ArgParser.FILE_ENCODING, 'utf-8')
        
    def test_getArguments_match(self):
        ArgParser.getArguments(['CAT', 'match:\\Atest\\Z'])
        self.assertCountEqual(ArgParser.FILE_MATCH, ['\\Atest\\Z'])
        ArgParser.getArguments(['CAT', 'match=\\Atest\\Z'])
        self.assertCountEqual(ArgParser.FILE_MATCH, ['\\Atest\\Z'] * 2)
        
    def test_getArguments_find(self):
        ArgParser.getArguments(['CAT', 'find=Test123'])
        self.assertCountEqual(ArgParser.FILE_SEARCH, ['Test123'])
        ArgParser.getArguments(['CAT', 'find:Test123'])
        self.assertCountEqual(ArgParser.FILE_SEARCH, ['Test123'] * 2)
        
    def test_getArguments_trunc(self):
        ArgParser.getArguments(['CAT', 'trunc=0:'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [0, None, None])
        ArgParser.getArguments(['CAT', 'trunc:1:'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [0, None, None])
        ArgParser.getArguments(['CAT', 'trunc:2:'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [1, None, None])
        ArgParser.getArguments(['CAT', 'trunc:::-1'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [None, None, -1])
        ArgParser.getArguments(['CAT', 'trunc:4::-5'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [3, None, -5])
        ArgParser.getArguments(['CAT', 'trunc:4:6:-5'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [3, 6, -5])
        ArgParser.getArguments(['CAT', 'trunc=:2*4-3:'])
        self.assertListEqual(ArgParser.FILE_TRUNCATE, [None, 5, None])
        
    def test_getArguments_cut(self):
        args, _, _, _, _ = ArgParser.getArguments(['CAT', '[2*5:20]'])
        self.assertCountEqual(args, [(ARGS_CUT, '[2*5:20]')])
        
    def test_getArguments_replace(self):
        args, _, _, _, _ = ArgParser.getArguments(['CAT', '[test,404]'])
        self.assertCountEqual(args, [(ARGS_REPLACE, '[test,404]')])
        
    def test_getArguments_dir(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', test_file_dir])
        self.assertCountEqual(args, [])
        self.assertCountEqual(unknown_args, [])
        self.assertGreaterEqual(len(known_files), 7)
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, [])
        
    def test_getArguments_files_equal_dir(self):
        _, _, known_files_dir, _, _ = ArgParser.getArguments(['CAT', test_file_dir])
        _, _, known_files_files, _, _ = ArgParser.getArguments(['CAT', test_file_dir + '/**.txt'])
        self.assertCountEqual(known_files_dir, known_files_files)
        
    def test_getArguments_unknown_file(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', 'testTesttest', 'test-file.txt'])
        self.assertCountEqual(args, [])
        self.assertCountEqual(unknown_args, [])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(echo_args, [])
        self.assertEqual(len(unknown_files), 2)
        self.assertIn('testTesttest', ''.join(unknown_files))
        self.assertIn('test-file.txt', ''.join(unknown_files))
        
    def test_getArguments_unknown_args(self):
        _, unknown_args, _, _, _ = ArgParser.getArguments(['CAT', '--test-file.txt'])
        self.assertCountEqual(unknown_args, ['--test-file.txt'])
        
    def test_getArguments_echo_args(self):
        args, unknown_args, known_files, unknown_files, echo_args = ArgParser.getArguments(['CAT', '-n', '-E', '-n', 'random', test_file_dir])
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, ['-n', '-E'])
        self.assertCountEqual(unknown_args, [])
        self.assertCountEqual(known_files, [])
        self.assertCountEqual(unknown_files, [])
        self.assertCountEqual(echo_args, ['-n', 'random', test_file_dir])
        
    def test_getArguments_echo_args_recursive(self):
        args = []
        echo = ArgParser.__addArgument__(args, [], [], [], '-nEn')
        args = list(map(lambda x: x[1], args))
        self.assertCountEqual(args, ['-n', '-E'])
        self.assertEqual(echo, True)


from tempfile import NamedTemporaryFile

class TmpFileHelper():
    def __init__(self) -> None:
        self.tmpFiles = []
        
    def getGeneratedTempFiles(self) -> list:
        return self.tmpFiles
        
    def generateTempFileName(self) -> str:
        tmp_file = NamedTemporaryFile(delete=False).name
        self.tmpFiles.append(tmp_file)
        return tmp_file


from unittest import TestCase
import os

from cat_win.util.checksum import getChecksumFromFile
# import sys
# sys.path.append('../cat_win')


test_file_path = os.path.join(os.path.dirname(__file__), 'texts', 'test.txt')


class TestChecksum(TestCase):
    def test_checksum(self):
        expected_output = ''
        expected_output += '\tCRC32:   C7222F64\n'
        expected_output += '\tMD5:     0f85f8d2b6783c06d40755ab54906862\n'
        expected_output += '\tSHA1:    1c50f01c5be5fc4795817e14f5deccbd51bf0934\n'
        expected_output += '\tSHA256:  dfb46efd198ad4b1204308c2be1c5e0254effe0de3ed314dea394cafdfd1749f\n'
        expected_output += '\tSHA512:  bf35256e790288a6dbf93b7327e1f97840349e0490b48848273663d34ad493e73a325df6725d7b0d5eb680d751bfe3cc49bd82ce9e8527412ce2cc8355a391a1\n'
        self.assertEqual(getChecksumFromFile(test_file_path), expected_output)
        
    def test_checksum_colored(self):
        expected_output = ''
        expected_output += '\tXCRC32:   C7222F64Y\n'
        expected_output += '\tXMD5:     0f85f8d2b6783c06d40755ab54906862Y\n'
        expected_output += '\tXSHA1:    1c50f01c5be5fc4795817e14f5deccbd51bf0934Y\n'
        expected_output += '\tXSHA256:  dfb46efd198ad4b1204308c2be1c5e0254effe0de3ed314dea394cafdfd1749fY\n'
        expected_output += '\tXSHA512:  bf35256e790288a6dbf93b7327e1f97840349e0490b48848273663d34ad493e73a325df6725d7b0d5eb680d751bfe3cc49bd82ce9e8527412ce2cc8355a391a1Y\n'
        self.assertEqual(getChecksumFromFile(test_file_path, ['X', 'Y']), expected_output)

# python -m unittest discover -s tests -p test*.py



from unittest import TestCase

from cat_win.util.FileAttributes import _convert_size
# import sys
# sys.path.append('../cat_win')


class TestFileAttributes(TestCase):
    def test__convert_size_zero(self):
        self.assertEqual(_convert_size(0), '0 B')
        
    def test__convert_size_EdgeKB(self):
        self.assertEqual(_convert_size(1023), '1023.0 B')
    
    def test__convert_size_KB_exact(self):
        self.assertEqual(_convert_size(1024), '1.0 KB')
        
    def test__convert_size_KB(self):
        self.assertEqual(_convert_size(1836), '1.79 KB')
        
    def test__convert_size_RoundKB(self):
        self.assertEqual(_convert_size(2044), '2.0 KB')
        
    def test__convert_size_TB(self):
        self.assertEqual(_convert_size(1024*1024*1024*1024), '1.0 TB')
        
    def test__convert_size_UnevenTB(self):
        self.assertEqual(_convert_size(1024*1024*1024*1024 * 2.3), '2.3 TB')
        
    def test__convert_size_YB(self):
        self.assertEqual(_convert_size(1024*1024*1024*1024*1024*1024*1024*1024), '1.0 YB')
        
    def test__convert_size_EdgeYB(self):
        self.assertEqual(_convert_size(1024*1024*1024*1024*1024*1024*1024*1024 * 1023.99), '1023.99 YB')
        
    def test__convert_size_OutOfRange(self):
        self.assertEqual(_convert_size(1024*1024*1024*1024*1024*1024*1024*1024*1024), '1.0 ?')



from unittest import TestCase

from cat_win.util.File import File
# import sys
# sys.path.append('../cat_win')


class TestFile(TestCase):
    def test_file_default(self):
        file = File('TestPath', 'TestName')
        
        self.assertEqual(file.containsQueried, False)
        self.assertEqual(file.path, 'TestPath')
        self.assertEqual(file.displayname, 'TestName')
        
    def test_file_queried(self):
        file = File('TestPath', 'TestName')
        self.assertEqual(file.containsQueried, False)
        
        file.setContainsQueried(False)
        self.assertEqual(file.containsQueried, False)
        
        file.setContainsQueried(True)
        self.assertEqual(file.containsQueried, True)
        
        file.setContainsQueried(False)
        self.assertEqual(file.containsQueried, True)

# python -m unittest discover -s tests -p test*.py




class ArgConstant():
    def __init__(self, shortForm, longForm, help, id):
        self.shortForm = shortForm
        self.longForm = longForm
        self.help = help
        self.id = id


ARGS_HELP, ARGS_NUMBER, ARGS_ENDS, ARGS_TABS, ARGS_SQUEEZE = range(0, 5)
ARGS_REVERSE, ARGS_COUNT, ARGS_BLANK, ARGS_FILES, ARGS_INTERACTIVE = range(5, 10)
ARGS_CLIP, ARGS_CHECKSUM, ARGS_DEC, ARGS_HEX, ARGS_BIN = range(10, 15)
ARGS_VERSION, ARGS_DEBUG, ARGS_CUT, ARGS_REPLACE, ARGS_DATA = range(15, 20)
ARGS_CONFIG, ARGS_LLENGTH, ARGS_ONELINE, ARGS_PEEK, ARGS_NOCOL = range(20, 25)
ARGS_EOF, ARGS_B64E, ARGS_B64D, ARGS_FFILES, ARGS_GREP = range(25, 30)
ARGS_NOBREAK, ARGS_ECHO, ARGS_CCOUNT, ARGS_HEXVIEW, ARGS_BINVIEW = range(30, 35)
ARGS_NOKEYWORD, ARGS_RECONFIGURE = range(35, 37)

ALL_ARGS = [[['-h', '--help'], 'show this help message and exit', ARGS_HELP],
            [['-v', '--version'], 'output version information and exit', ARGS_VERSION],
            [['-d', '--debug'], 'show debug information', ARGS_DEBUG],
            [['-n', '--number'], 'number all output lines', ARGS_NUMBER],
            [['-l', '--linelength'], 'display the length of each line', ARGS_LLENGTH],
            [['-e', '--ends'], 'display $ at the end of each line', ARGS_ENDS],
            [['-t', '--tabs'], 'display TAB characters as ^I', ARGS_TABS],
            [['--eof', '--eof'], 'display EOF characters as ^EOF', ARGS_EOF],
            [['-u', '--unique'], 'suppress repeated output lines', ARGS_SQUEEZE],
            [['-b', '--blank'], 'hide empty lines', ARGS_BLANK],
            [['-r', '--reverse'], 'reverse output', ARGS_REVERSE],
            [['-p', '--peek'], 'only print the first and last lines', ARGS_PEEK],
            [['-s', '--sum'], 'show sum of lines', ARGS_COUNT],
            [['-S', '--SUM'], 'ONLY show sum of lines', ARGS_CCOUNT],
            [['-f', '--files'], 'list applied files', ARGS_FILES],
            [['-F', '--FILES'], 'ONLY list applied files and file sizes', ARGS_FFILES],
            [['-g', '--grep'], 'only show lines containing queried keywords', ARGS_GREP],
            [['-i', '--interactive'], 'use stdin', ARGS_INTERACTIVE],
            [['-o', '--oneline'], 'take only the first stdin-line', ARGS_ONELINE],
            [['-E', '--ECHO'], 'handle every following parameter as stdin', ARGS_ECHO],
            [['-c', '--clip'], 'copy output to clipboard', ARGS_CLIP],
            [['-m', '--checksum'], 'show the checksums of all files', ARGS_CHECKSUM],
            [['-a', '--attributes'], 'show meta-information about the files', ARGS_DATA],
            [['--dec', '--DEC'], 'convert decimal numbers to hexadecimal and binary', ARGS_DEC],
            [['--hex', '--HEX'], 'convert hexadecimal numbers to decimal and binary', ARGS_HEX],
            [['--bin', '--BIN'], 'convert binary numbers to decimal and hexadecimal', ARGS_BIN],
            [['--b64e', '--b64e'], 'encode the input to base64', ARGS_B64E],
            [['--b64d', '--b64d'], 'decode the input from base64', ARGS_B64D],
            [['--hexview', '--HEXVIEW'], 'display the raw byte representation in hexadecimal', ARGS_HEXVIEW],
            [['--binview', '--binview'], 'display the raw byte representation in binary', ARGS_BINVIEW],
            [['--nc', '--nocolor'], 'disable colored output', ARGS_NOCOL],
            [['--nb', '--nobreak'], 'do not interrupt the output on queried keywords', ARGS_NOBREAK],
            [['--nk', '--nokeyword'], 'inverse the grep output', ARGS_NOKEYWORD],
            [['-R', '--reconfigure'], 'reconfigure the stdin and stdout with the parsed encoding', ARGS_RECONFIGURE],
            [['--config', '--config'], 'change color configuration', ARGS_CONFIG]]

ALL_ARGS = [ArgConstant(x[0][0], x[0][1], x[1], x[2]) for x in ALL_ARGS]
HIGHEST_ARG_ID = max([m.id for m in ALL_ARGS])



"""Simple OS Independent 'cat' Command-line Tool made in Python."""

__project__ = 'cat_win'
__version__ = '1.4.0'
__sysversion__ = '3.11.2 (main, Feb  7 2023, 16:38:35) [MSC v.1934 64 bit (AMD64)]'
__author__ = 'Silas A. Kraume'
__url__ = 'https://github.com/SilenZcience/cat_win'



from functools import lru_cache
from heapq import nlargest

from cat_win.const.ArgConstants import *
from cat_win.util.File import File


class Holder():
    def __init__(self) -> None:
        self.files = []  # all files, including tmp-file from stdin
        self._inner_files = []
        self.args = []  # list of all used parameters: format [[id, param]]
        self.args_id = [False] * (HIGHEST_ARG_ID + 1)
        self.temp_file_stdin = None  # if stdin is used, this temp_file will contain the stdin-input
        self.temp_file_echo = None  # if ARGS_ECHO is used, this temp_file will contain the following parameters
        self.reversed = False
        
        # the amount of chars neccessary to display the last file
        self.fileNumberPlaceHolder = 0
        # the sum of all lines of all files
        self.allFilesLinesSum = 0
        # the sum of lines for each file individually
        self.allFilesLines = {}
        # the amount of chars neccessary to display the last line (breaks on base64 decoding)
        self.fileLineNumberPlaceHolder = 0
        # the amount of chars neccessary to display the longest line within all files (breaks on base64 decoding)
        self.fileLineLengthPlaceHolder = 0

        self.clipBoard = ''
    
    def _getFileDisplayName(self, file: str) -> str:
        """
        return the display name of a file. Expects self.temp_file_stdin
        and self.temp_file_echo to be set already.
        
        Parameters_
        file (str):
            a path of a file
            
        Returns:
        (str):
            the display name for the file. Either the path itself
            or a special identifier von stdin or echo inputs
        """
        if file == self.temp_file_stdin:
            return '<STDIN>'
        elif file == self.temp_file_echo:
            return '<ECHO>'
        return file
    
    def setFiles(self, files: list) -> None:
        self.files = [File(path, self._getFileDisplayName(path)) for path in files]
        self._inner_files = files[:]

    def setArgs(self, args: list) -> None:
        self.args = args
        for id, _ in self.args:
            self.args_id[id] = True
        if self.args_id[ARGS_B64E]:
            self.args_id[ARGS_NOCOL] = True
            # prefix will be deleted anyway
            self.args_id[ARGS_LLENGTH] = False
            self.args_id[ARGS_NUMBER] = False
        self.reversed = self.args_id[ARGS_REVERSE]

    def setTempFileStdIn(self, file: str) -> None:
        self.temp_file_stdin = file
        
    def setTempFileEcho(self, file: str) -> None:
        self.temp_file_echo = file
    
    def __calcFileNumberPlaceHolder__(self) -> None:
        self.fileNumberPlaceHolder = len(str(len(self.files)))

    def __count_generator__(self, reader):
        """
        Parameters:
        reader (method):
            the method to read from
        
        Yields:
        b (bytes):
            the bytes in chunks read from the reader
        """
        b = reader(1024 * 1024)
        while b:
            yield b
            b = reader(1024 * 1024)

    @lru_cache(maxsize=10)
    def __getFileLinesSum__(self, file: str) -> int:
        with open(file, 'rb') as fp:
            c_generator = self.__count_generator__(fp.raw.read)
            linesSum = sum(buffer.count(b'\n') for buffer in c_generator) + 1
        return linesSum

    def __calcPlaceHolder__(self) -> None:
        fileLines = []
        for file in self._inner_files:
            fileLineSum = self.__getFileLinesSum__(file)
            fileLines.append(fileLineSum)
            self.allFilesLines[file] = fileLineSum
        self.allFilesLinesSum = sum(fileLines)
        self.fileLineNumberPlaceHolder = len(str(max(fileLines)))

    @lru_cache(maxsize=10)
    def __calcMaxLineLength__(self, file: str) -> int:
        """
        Calculate self.fileLineLengthPlaceHolder for a single file.
        
        Parameters:
        file (str):
            a string representation of a file (-path)
            
        Returns:
        (int):
            the length of the placeholder to represent
            the longest line within the file
        """
        heap = []
        lines = []
        with open(file, 'rb') as fp:
            lines = fp.readlines()
        
        heap = nlargest(1, lines, len)
        if len(heap) == 0:
            return 0
        # also check the longest line against the last line because
        # the lines still contain (\r)\n, except the last line does not
        longest_line_len = len(heap[0][:-1].rstrip(b'\n').rstrip(b'\r'))
        last_line_len = len(lines[-1].rstrip(b'\n').rstrip(b'\r'))
        
        return len(str(max(longest_line_len, last_line_len)))

    def __calcFileLineLengthPlaceHolder__(self) -> None:
        self.fileLineLengthPlaceHolder = max(self.__calcMaxLineLength__(file)
                                             for file in self._inner_files)
        
    def setDecodingTempFiles(self, temp_files: list) -> None:
        self._inner_files = temp_files[:]

    def generateValues(self, encoding: str) -> None:
        self.__calcFileNumberPlaceHolder__()
        if self.args_id[ARGS_B64D]:
            from cat_win.util.Base64 import _decodeBase64
            for i, file in enumerate(self.files):
                with open(file.path, 'rb') as fp:
                    with open(self._inner_files[i], 'wb') as f:
                        f.write(_decodeBase64(fp.read().decode(encoding)))
        self.__calcPlaceHolder__()
        self.__calcFileLineLengthPlaceHolder__()



from re import finditer

from cat_win.const.ColorConstants import C_KW


class StringFinder:
    def __init__(self, literals: list, regex: list) -> None:
        self.kw_literals = literals
        self.kw_regex = regex

    def _findliterals(self, sub: str, s: str):
        """
        Generate lists containing the position of sub in s.
        
        Parameters:
        sub (str):
            the substring to search for
        s (str):
            the string to search in
        
        Yields:
        (list):
            containing the start and end indeces like [start, end]
        """
        l = len(sub)
        i = s.find(sub)
        while i != -1:
            yield [i, i+l]
            i = s.find(sub, i+1)

    def _findregex(self, pattern: str, s: str):
        """
        Generate lists containing the position of pattern in s.
        
        Parameters:
        pattern (str):
            the regex pattern to search for
        s (str):
            the string to search in
        
        Yields:
        (list):
            containing the start and end indeces like [start, end]
        """
        for match in finditer(fr'{pattern}', s):
            yield list(match.span())

    def _optimizeIntervals(self, intervals: list) -> list:
        """
        optimize/shorten/merge overlapping intervalls for partially
        color encoded lines. Needed when multiple
        search-keywords apply to the same line.
        
        Parameters:
        intervals (list):
            the intervals to optimize like [[start, end], ...]
            
        Returns:
        stack (list)
            the optimized intervals
        """
        if not intervals:
            return []
        intervals.sort()
        stack = []
        stack.append(intervals[0])
        for interval in intervals:
            if stack[-1][0] <= interval[0] <= stack[-1][-1]:
                stack[-1][-1] = max(stack[-1][-1], interval[-1])
            else:
                stack.append(interval)
        return stack

    def _mergeKeywordIntervals(self, fList: list, mList: list) -> list:
        fList = self._optimizeIntervals(fList)
        mList = self._optimizeIntervals(mList)
        kwList = []
        for f in fList:
            kwList.append([f[0], C_KW.FOUND])
            kwList.append([f[1], C_KW.RESET_FOUND])
        for rf in mList:
            kwList.append([rf[0], C_KW.MATCHED])
            kwList.append([rf[1], C_KW.RESET_MATCHED])
        kwList.sort(reverse=True)
        return kwList

    def findKeywords(self, line: str) -> tuple:
        """
        calculate the positions for ANSI-ColorCodes.
        
        Parameters:
        line (str):
            the string in which to search for all occurences of
            self.kw_literals and self.kw_regex
            
        Returns:
        (tuple):
            containing three lists:
                the index positions of all intervals
                the found substrings
                the matched patterns
        """
        found_list = []
        found_position = []
        matched_list = []
        matched_position = []

        for keyword in self.kw_literals:
            for f in self._findliterals(keyword, line):
                found_position.append(f)
                found_list.append((keyword, f))

        for keyword in self.kw_regex:
            for m in self._findregex(keyword, line):
                matched_position.append(m)
                matched_list.append((keyword, m))

        return (self._mergeKeywordIntervals(found_position, matched_position), found_list, matched_list)



from configparser import ConfigParser
from os.path import join as path_join

from cat_win.const.ColorConstants import ColorOptions, C_KW


class Config:
    default_dic = {C_KW.NUMBER: ColorOptions.Fore['GREEN'],
                   C_KW.LINE_LENGTH: ColorOptions.Fore['LIGHTBLUE'],
                   C_KW.ENDS: ColorOptions.Back['YELLOW'],
                   C_KW.TABS: ColorOptions.Back['YELLOW'],
                   C_KW.CONVERSION: ColorOptions.Fore['CYAN'],
                   C_KW.REPLACE: ColorOptions.Fore['YELLOW'],
                   C_KW.FOUND: ColorOptions.Fore['RED'],
                   C_KW.FOUND_MESSAGE: ColorOptions.Fore['MAGENTA'],
                   C_KW.MATCHED: ColorOptions.Back['CYAN'],
                   C_KW.MATCHED_MESSAGE: ColorOptions.Fore['LIGHTCYAN'],
                   C_KW.CHECKSUM: ColorOptions.Fore['CYAN'],
                   C_KW.COUNT_AND_FILES: ColorOptions.Fore['CYAN'],
                   C_KW.ATTRIB_POSITIVE: ColorOptions.Fore['LIGHTGREEN'],
                   C_KW.ATTRIB_NEGATIVE: ColorOptions.Fore['LIGHTRED'],
                   C_KW.ATTRIB: ColorOptions.Fore['CYAN'],
                   C_KW.MESSAGE_INFORMATION: ColorOptions.Fore['LIGHTBLACK'],
                   C_KW.MESSAGE_IMPORTANT: ColorOptions.Fore['YELLOW'],
                   C_KW.MESSAGE_WARNING: ColorOptions.Fore['RED'],
                   C_KW.RAWVIEWER: ColorOptions.Fore['LIGHTBLACK']}
    elements = list(default_dic.keys())

    def __init__(self, workingDir: str) -> None:
        """
        Initialise the Config() object to load and save
        the color configs.
        
        Parameters:
        workingDir (str):
            the working directory path of the package
        """
        self.workingDir = workingDir
        self.configFile = path_join(self.workingDir, 'cat.config')
        
        self.exclusive_definitions = {'Fore': [C_KW.FOUND],  # can only be Foreground
                                      'Back': [C_KW.MATCHED]}  # can only be Background
        self.configParser = ConfigParser()
        self.color_dic = {}
        
        self.longestCharCount = 30
        self.ROWS = 3

    def loadConfig(self) -> dict:
        """
        Load the Color Configuration from the config file.
        
        Returns:
        color_dic (dict):
            a dictionary translating from C_KW-keywords to ANSI-Colorcodes
        On Error: Return the default color config
        """
        try:
            self.configParser.read(self.configFile)
            configColors = self.configParser['COLORS']
            for element in self.elements:
                try:
                    type, color = configColors[element].split('.')
                    self.color_dic[element] = (
                        ColorOptions.Fore[color] if type == 'Fore' else ColorOptions.Back[color]
                        )
                except KeyError:
                    self.color_dic[element] = self.default_dic[element]
        except KeyError:
            self.configParser['COLORS'] = {}
            # If an error occures we simplfy use the default colors
            self.color_dic = self.default_dic

        # The Reset Codes should always be the same
        self.color_dic[C_KW.RESET_ALL] = ColorOptions.Style['RESET']
        self.color_dic[C_KW.RESET_FOUND] = ColorOptions.Fore['RESET']
        self.color_dic[C_KW.RESET_MATCHED] = ColorOptions.Back['RESET']

        return self.color_dic

    def _printGetAllAvailableColors(self) -> list:
        """
        prints all available color options to choose from.
        
        Returns:
        options (list):
            the same list containing all available colors.
        """
        print('Here is a list of all available color options you may choose:')
        
        ForeOptions = list(ColorOptions.Fore.items())
        ForeOptions = [(k, v) for k, v in ForeOptions if k != 'RESET']
        BackOptions = list(ColorOptions.Back.items())
        BackOptions = [(k, v) for k, v in BackOptions if k != 'RESET']
        indexOffset = len(str(len(ForeOptions) + len(BackOptions) + 1))
        
        configMenu = ''
        options = []
        
        for index in range(len(ForeOptions)):
            key, value = ForeOptions[index]
            coloredOption = f"{value}Fore.{key}{ColorOptions.Style['RESET']}"
            configMenu += f"{index+1: <{indexOffset}}: {coloredOption: <{self.longestCharCount+len(value)}}"
            if index % self.ROWS == self.ROWS-1:
                configMenu += '\n'
            options.append('Fore.' + key)
        configMenu += '\n'
        for index in range(len(BackOptions)):
            key, value = BackOptions[index]
            coloredOption = f"{value}Back.{key}{ColorOptions.Style['RESET']}"
            configMenu += f"{len(ForeOptions)+index+1: <{indexOffset}}: {coloredOption: <{self.longestCharCount+len(value)}}"
            if index % self.ROWS == self.ROWS-1:
                configMenu += '\n'
            options.append('Back.' + key)
        configMenu += '\n'
        
        print(configMenu)
        return options

    def _printAllAvailableElements(self) -> None:
        """
        print all available elements which color can be changed.
        """
        print('Here is a list of all available elements you may change:')
        
        self.longestCharCount = max(map(len, self.elements)) + 12
        indexOffset = len(str(len(self.elements) + 1))

        configMenu = ''
        for index in range(len(self.elements)):
            element = self.elements[index]
            coloredElement = f"{self.color_dic[element]}{element}{ColorOptions.Style['RESET']}"
            configMenu += f"{index+1: <{indexOffset}}: {coloredElement: <{self.longestCharCount+len(self.color_dic[element])}}"
            if index % self.ROWS == self.ROWS-1:
                configMenu += '\n'
        
        print(configMenu)

    def saveConfig(self) -> None:
        """
        Guide the User through the configuration options and save the changes.
        Assume, that the current config is already loaded/the method loadConfig() was already called.
        """
        self._printAllAvailableElements()
        keyword = ''
        while (not keyword in self.elements):
            if keyword != '':
                print(f"Something went wrong. Unknown keyword '{keyword}'")
            keyword = input('Input name of keyword to change: ')
            if keyword.isdigit():
                keyword = self.elements[int(keyword)-1] if (
                    0 < int(keyword) <= len(self.elements)) else keyword
        print(f"Successfully selected element '{keyword}'.")

        color_options = self._printGetAllAvailableColors()
        color = ''
        while (not color in color_options):
            if color != '':
                print(f"Something went wrong. Unknown option '{color}'.")
            color = input('Input color: ')
            if color.isdigit():
                color = color_options[int(color)-1] if (
                    0 < int(color) <= len(color_options)) else color

        if keyword in self.exclusive_definitions['Fore'] and color.startswith('Back'):
            print(f"An Error occured: '{keyword}' can only be of style 'Fore'")
            return
        if keyword in self.exclusive_definitions['Back'] and color.startswith('Fore'):
            print(f"An Error occured: '{keyword}' can only be of style 'Back'")
            return

        print(f"Successfully selected element '{color}'.")

        self.configParser['COLORS'][keyword] = color
        try:
            with open(self.configFile, 'w') as conf:
                self.configParser.write(conf)
            print(f"Successfully updated config file:\n\t{self.configFile}")
        except OSError:
            print(f"Could not write to config file:\n\t{self.configFile}")



from sys import exit

try:
    import cat_win.cat as cat
except KeyboardInterrupt:
    exit(1)
except Exception as e:
    print('an error occured while loading the module:')
    print(e)
    exit(1)

def shell_entry_point():
    cat.shell_main()

def entry_point():
    cat.main()

if __name__ == '__main__':
    entry_point()



ESC_CODE = '\x1b'


def idToCode(code: int) -> str:
    return f"{ESC_CODE}[{code}m"


class ColorOptions:
    __Fore = {
        'RESET': 39,
            
        'BLACK':   30,
        'RED':     31,
        'GREEN':   32,
        'YELLOW':  33,
        'BLUE':    34,
        'MAGENTA': 35,
        'CYAN':    36,
        'WHITE':   37,
        
        'LIGHTBLACK':   90,
        'LIGHTRED':     91,
        'LIGHTGREEN':   92,
        'LIGHTYELLOW':  93,
        'LIGHTBLUE':    94,
        'LIGHTMAGENTA': 95,
        'LIGHTCYAN':    96,
        'LIGHTWHITE':   97
        }
    __Back = {
        'RESET': 49,
            
        'BLACK':   40,
        'RED':     41,
        'GREEN':   42,
        'YELLOW':  43,
        'BLUE':    44,
        'MAGENTA': 45,
        'CYAN':    46,
        'WHITE':   47,
          
        'LIGHTBLACK':   100,
        'LIGHTRED':     101,
        'LIGHTGREEN':   102,
        'LIGHTYELLOW':  103,
        'LIGHTBLUE':    104,
        'LIGHTMAGENTA': 105,
        'LIGHTCYAN':    106,
        'LIGHTWHITE':   107
        }
    __Style = {
        'RESET': 0
        }

    Fore  = {'NONE': ''}
    Back  = {'NONE': ''}
    Style = {'NONE': ''}
    
    for key in __Fore:
        Fore[key]  = idToCode(__Fore[key])
    for key in __Back:
        Back[key]  = idToCode(__Back[key])
    for key in __Style:
        Style[key] = idToCode(__Style[key])


class C_KW:
    RESET_ALL = 'reset_all'
    RESET_FOUND = 'reset_found'
    RESET_MATCHED = 'reset_matched'

    NUMBER = 'line_numbers'
    LINE_LENGTH = 'line_length'
    ENDS = 'line_ends'
    TABS = 'tab_characters'
    CONVERSION = 'number_conversion'
    REPLACE = 'substring_replacement'
    FOUND = 'found_keyword'
    FOUND_MESSAGE = 'found_keyword_message'
    MATCHED = 'matched_pattern'
    MATCHED_MESSAGE = 'matched_pattern_message'
    CHECKSUM = 'checksum_message'
    COUNT_AND_FILES = 'processed_message'
    ATTRIB = 'file_attribute_message'
    ATTRIB_POSITIVE = 'active_file_attributes'
    ATTRIB_NEGATIVE = 'missing_file_attributes'
    
    MESSAGE_INFORMATION = 'message_information'
    MESSAGE_IMPORTANT = 'message_important'
    MESSAGE_WARNING = 'message_warning'
    
    RAWVIEWER = 'raw_viewer'



from zlib import crc32 as crc32_hash
import hashlib


def getChecksumFromFile(file: str, colors = None) -> str:
    """
    Calculates and returns the CRC32, MD5, SHA1, SHA256, SHA512
    hashes of a file.
    
    Parameters:
    file (str):
        a string representation of a file (-path)
    colors (list):
        a list with 2 elements like [COLOR_CHECKSUM, COLOR_RESET]
        containing the ANSI-Colorcodes used in the returned string.
    
    Returns:
    checksum (str):
        a formatted string representation of all checksums calculated
    """
    if colors == None or len(colors) < 2:
        colors = ['', '']
    BUF_SIZE = 65536  # 64kb
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()
    sha512 = hashlib.sha512()
    crc32 = 0

    with open(file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
            sha1.update(data)
            sha256.update(data)
            sha512.update(data)
            crc32 = crc32_hash(data, crc32)

    crc32 = f"{(crc32 & 0xFFFFFFFF):08X}"
    
    checksum =  f"\t{colors[0]}{'CRC32:' : <9}{str(crc32)}{colors[1]}\n"
    checksum += f"\t{colors[0]}{'MD5:'   : <9}{str(md5.hexdigest())}{colors[1]}\n"
    checksum += f"\t{colors[0]}{'SHA1:'  : <9}{str(sha1.hexdigest())}{colors[1]}\n"
    checksum += f"\t{colors[0]}{'SHA256:': <9}{str(sha256.hexdigest())}{colors[1]}\n"
    checksum += f"\t{colors[0]}{'SHA512:': <9}{str(sha512.hexdigest())}{colors[1]}\n"
    return checksum



from unittest import TestCase
import os

from cat_win.const.ArgConstants import *
from cat_win.util.Holder import Holder
# import sys
# sys.path.append('../cat_win')

test_file_dir = os.path.join(os.path.dirname(__file__), 'texts')
test_file_path        = os.path.join(test_file_dir, 'test.txt')
test_file_edge_case_1 = os.path.join(test_file_dir, 'test_holderEdgeCase_1.txt')
test_file_edge_case_2 = os.path.join(test_file_dir, 'test_holderEdgeCase_2.txt')
test_file_edge_case_3 = os.path.join(test_file_dir, 'test_holderEdgeCase_3.txt')
test_file_edge_case_4 = os.path.join(test_file_dir, 'test_holderEdgeCase_4.txt')
test_file_empty       = os.path.join(test_file_dir, 'test_empty.txt')
holder = Holder()


class TestHolder(TestCase):
    def test__calcFileLineLengthPlaceHolder__(self):
        holder.setFiles([test_file_path])
        holder.__calcFileLineLengthPlaceHolder__()
        self.assertEqual(holder.fileLineLengthPlaceHolder, 2)

    def test__calcFileLineLengthPlaceHolder__Edge(self):
        holder.setFiles([test_file_edge_case_1])
        holder.__calcFileLineLengthPlaceHolder__()
        self.assertEqual(holder.fileLineLengthPlaceHolder, 1)

        holder.setFiles([test_file_edge_case_2])
        holder.__calcFileLineLengthPlaceHolder__()
        self.assertEqual(holder.fileLineLengthPlaceHolder, 2)
        
        holder.setFiles([test_file_edge_case_3, test_file_edge_case_4])
        holder.__calcFileLineLengthPlaceHolder__()
        self.assertEqual(holder.fileLineLengthPlaceHolder, 1)
        
        holder.setFiles([test_file_edge_case_2, test_file_edge_case_4])
        holder.__calcFileLineLengthPlaceHolder__()
        self.assertEqual(holder.fileLineLengthPlaceHolder, 2)
        
    def test___calcMaxLineLength___empty(self):
        self.assertEqual(holder.__calcMaxLineLength__(test_file_empty), 0)

    def test_allFilesLinesSum(self):
        holder.setFiles([test_file_path, test_file_edge_case_1])
        holder.__calcPlaceHolder__()
        self.assertEqual(holder.allFilesLinesSum, 10)

        holder.setFiles([test_file_path] * 13)
        holder.__calcPlaceHolder__()
        self.assertEqual(holder.allFilesLinesSum, 104)

    def test_fileLineNumberPlaceHolder(self):
        holder.setFiles([test_file_path])
        holder.__calcPlaceHolder__()
        self.assertEqual(holder.fileLineNumberPlaceHolder, 1)

        holder.setFiles([test_file_path, test_file_path])
        holder.__calcPlaceHolder__()
        self.assertEqual(holder.fileLineNumberPlaceHolder, 1)

        holder.setFiles([test_file_edge_case_3])
        holder.__calcPlaceHolder__()
        self.assertEqual(holder.fileLineNumberPlaceHolder, 2)

    def test_fileNumberPlaceHolder(self):
        holder.setFiles([test_file_edge_case_1] * 9)
        holder.__calcFileNumberPlaceHolder__()
        self.assertEqual(holder.fileNumberPlaceHolder, 1)

        holder.setFiles([test_file_edge_case_1] * 10)
        holder.__calcFileNumberPlaceHolder__()
        self.assertEqual(holder.fileNumberPlaceHolder, 2)

        holder.setFiles([test_file_edge_case_1] * 99)
        holder.__calcFileNumberPlaceHolder__()
        self.assertEqual(holder.fileNumberPlaceHolder, 2)

        holder.setFiles([test_file_edge_case_1] * 100)
        holder.__calcFileNumberPlaceHolder__()
        self.assertEqual(holder.fileNumberPlaceHolder, 3)
        
    def test_setArgBase64(self):
        holder.setArgs([(ARGS_B64E, '--b64e')])
        self.assertEqual(holder.args_id[ARGS_B64E], True)
        self.assertEqual(holder.args_id[ARGS_NOCOL], True)
        self.assertEqual(holder.args_id[ARGS_LLENGTH], False)
        self.assertEqual(holder.args_id[ARGS_NUMBER], False)

    def test__getFileDisplayName(self):
        holder.setTempFileStdIn('STDINFILE')
        holder.setTempFileEcho('TEMPFILEECHO')
        holder.setFiles([test_file_edge_case_3, 'STDINFILE', test_file_edge_case_4, 'TEMPFILEECHO'])
        
        self.assertEqual(holder._getFileDisplayName('STDINFILE'), '<STDIN>')
        self.assertEqual(holder._getFileDisplayName('TEMPFILEECHO'), '<ECHO>')
        self.assertEqual(holder._getFileDisplayName(test_file_edge_case_3), test_file_edge_case_3)
        self.assertEqual(holder._getFileDisplayName(test_file_edge_case_4), test_file_edge_case_4)


from io import StringIO
from unittest import TestCase
from unittest.mock import patch
import os

from cat_win.util.Holder import Holder
import cat_win.cat as cat
# import sys
# sys.path.append('../cat_win')


test_file_path = os.path.join(os.path.dirname(__file__), 'texts', 'test.txt')
test_file_content = []
with open(test_file_path, 'r', encoding='utf-8') as f:
    test_file_content = f.read().split('\n')


holder = Holder()

@patch('cat_win.cat.holder', holder)
@patch('cat_win.cat.color_dic', dict.fromkeys(cat.color_dic, ''))
class TestCat(TestCase):
    maxDiff = None

    def tearDown(self):
        cat._CalculateLinePrefixSpacing.cache_clear()
        cat._CalculateLineLengthPrefixSpacing.cache_clear()
    
    def test_cat_output_default_file(self):
        holder.setFiles([test_file_path])
        holder.setArgs([])
        
        check_against = '\n'.join(test_file_content) + '\n'

        with patch('sys.stdout', new=StringIO()) as fake_out:
            cat.editFiles()
            self.assertEqual(fake_out.getvalue(), check_against)

    def test_cat_output_multiple_files(self):
        holder.setFiles([test_file_path, test_file_path, test_file_path])
        holder.setArgs([])

        check_against = '\n'.join(test_file_content * 3) + '\n'

        with patch('sys.stdout', new=StringIO()) as fake_out:
            cat.editFiles()
            self.assertEqual(fake_out.getvalue(), check_against)

    def test_cat_output_reverse(self):
        holder.setFiles([test_file_path])
        holder.setArgs([[5, '']]) #reverse

        check_against = test_file_content
        check_against.reverse()
        check_against = '\n'.join(check_against) + '\n'

        with patch('sys.stdout', new=StringIO()) as fake_out:
            cat.editFiles()
            self.assertEqual(fake_out.getvalue(), check_against)

    def test_cat_output_ends_and_tabs(self):
        holder.setFiles([test_file_path])
        holder.setArgs([[2, ''], [3, '']]) #ends & tabs

        check_against = ('\n'.join([c.replace('\t', '^I') + '$' for c in test_file_content]) +
                         '\n')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            cat.editFiles()
            self.assertEqual(fake_out.getvalue(), check_against)
            
    def test_cat__getLinePrefix_file_excess(self):
        holder.fileLineNumberPlaceHolder = 5
        self.assertEqual(cat._getLinePrefix(9, 1), '    9) ')
        
    def test_cat__getLinePrefix_file_occupied(self):
        holder.fileLineNumberPlaceHolder = 2
        self.assertEqual(cat._getLinePrefix(10, 1), '10) ')
    
    def test_cat__getLinePrefix_file_excess_long(self):
        holder.fileLineNumberPlaceHolder = 12
        self.assertEqual(cat._getLinePrefix(34719, 1), '       34719) ')
    
    def test_cat__getLinePrefix_file_occupied_long(self):
        holder.fileLineNumberPlaceHolder = 5
        self.assertEqual(cat._getLinePrefix(34718, 1), '34718) ')
        
    def test_cat__getLinePrefix_files_excess(self):
        holder.fileLineNumberPlaceHolder = 5
        holder.fileNumberPlaceHolder = 4
        holder.files = [1,2]
        self.assertEqual(cat._getLinePrefix(9, 1), '   1.    9) ')
        
    def test_cat__getLinePrefix_files_occupied(self):
        holder.fileLineNumberPlaceHolder = 3
        holder.fileNumberPlaceHolder = 2
        holder.files = [1,2]
        self.assertEqual(cat._getLinePrefix(987, 10), '10.987) ')
        
    def test_cat__getLinePrefix_files_excess_long(self):
        holder.fileLineNumberPlaceHolder = 12
        holder.fileNumberPlaceHolder = 10
        holder.files = [1,2]
        self.assertEqual(cat._getLinePrefix(101, 404), '       404.         101) ')
        
    def test_cat__getLinePrefix_files_occupied_long(self):
        holder.fileLineNumberPlaceHolder = 11
        holder.fileNumberPlaceHolder = 9
        holder.files = [1,2]
        self.assertEqual(cat._getLinePrefix(12345123451, 123456789), '123456789.12345123451) ')
    
    def test_cat__getLineLengthPrefix_string_excess(self):
        holder.fileLineLengthPlaceHolder = 5
        holder.setArgs([])
        self.assertEqual(cat._getLineLengthPrefix('testtest', 'abcdefghi'), 'testtest[    9] ')
    
    def test_cat__getLineLengthPrefix_string_occupied(self):
        holder.fileLineLengthPlaceHolder = 2
        holder.setArgs([])
        self.assertEqual(cat._getLineLengthPrefix('prefix', 'abcdefghij'), 'prefix[10] ')
        
    def test_cat__getLineLengthPrefix_bytes_excess(self):
        holder.fileLineLengthPlaceHolder = 5
        holder.setArgs([])
        self.assertEqual(cat._getLineLengthPrefix('testtest', b'abcdefghi'), 'testtest[    9] ')
    
    def test_cat__getLineLengthPrefix_bytes_occupied(self):
        holder.fileLineLengthPlaceHolder = 2
        holder.setArgs([])
        self.assertEqual(cat._getLineLengthPrefix('prefix', b'abcdefghij'), 'prefix[10] ')
        
    def test_removeAnsiCodesFromLine(self):
        red = '\x1b[31m'
        reset = '\x1b[0m'
        random_string = f"abc{red}defghij{reset}klmnopq{red}r{reset}"
        expected_output = 'abcdefghijklmnopqr'
        self.assertEqual(cat.removeAnsiCodesFromLine(random_string), expected_output)

    # def test_removeAnsiCodes(self):
    #     red_bold_underlined = '\x1b[31;1;4m'
    #     reset = '\x1b[0m'
    #     random_list = [(f"12{red_bold_underlined}34{reset}5', f'abc{red_bold_underlined}defghij{reset}klmnopq{red_bold_underlined}r{reset}")]
    #     expected_output = [('12345', 'abcdefghijklmnopqr')]
    #     self.assertEqual(cat.removeAnsiCodes(random_list), expected_output)

# python -m unittest discover -s cat_win.tests -p test*.py



from unittest import TestCase

from cat_win.util.Base64 import encodeBase64, decodeBase64
# import sys
# sys.path.append('../cat_win')


class TestBase64(TestCase):
    def test_encodeBase64(self):
        test_input = [('', 'Test'),
                      ('', '123404'),
                      ('', 'ÄÖÜ  TEST')]
        expected_output = [('', 'VGVzdAoxMjM0MDQKw4TDlsOcICBURVNU')]
        self.assertEqual(encodeBase64(test_input), expected_output)

    def test_decodeBase64(self):
        test_input = [('', 'VGVzdAoxMjM0MDQKw4TDlsOcICBURVNU')]
        expected_output = [('', 'Test'),
                           ('', '123404'),
                           ('', 'ÄÖÜ  TEST')]
        self.assertEqual(decodeBase64(test_input), expected_output)

    def test_decodeBase64_lineBreak(self):
        test_input = [('', 'VGVzdAoxMjM0MDQKw4'),
                      ('', 'TDlsOcICBURVNU')]
        expected_output = [('', 'Test'),
                           ('', '123404'),
                           ('', 'ÄÖÜ  TEST')]
        self.assertEqual(decodeBase64(test_input), expected_output)

# python -m unittest discover -s tests -p test*.py



from unittest import TestCase

from cat_win.util.Converter import Converter
# import sys
# sys.path.append('../cat_win')


converter = Converter()


class TestConverter(TestCase):
    def test_converter_dec(self):
        expected_output = '{Hexadecimal: 0x3039; Binary: 0b11000000111001}'
        self.assertEqual(converter._fromDEC(12345, True), expected_output)

        expected_output = '{Hexadecimal: 3039; Binary: 11000000111001}'
        self.assertEqual(converter._fromDEC(12345, False), expected_output)

    def test_converter_hex(self):
        expected_output = '{Decimal: 12345; Binary: 0b11000000111001}'
        self.assertEqual(converter._fromHEX('0x3039', True), expected_output)

        expected_output = '{Decimal: 12345; Binary: 11000000111001}'
        self.assertEqual(converter._fromHEX('0x3039', False), expected_output)

        expected_output = '{Decimal: 12345; Binary: 0b11000000111001}'
        self.assertEqual(converter._fromHEX('3039', True), expected_output)

        expected_output = '{Decimal: 12345; Binary: 11000000111001}'
        self.assertEqual(converter._fromHEX('3039', False), expected_output)

    def test_converter_bin(self):
        expected_output = '{Decimal: 12345; Hexadecimal: 0x3039}'
        self.assertEqual(converter._fromBIN(
            '0b11000000111001', True), expected_output)

        expected_output = '{Decimal: 12345; Hexadecimal: 3039}'
        self.assertEqual(converter._fromBIN(
            '0b11000000111001', False), expected_output)

        expected_output = '{Decimal: 12345; Hexadecimal: 0x3039}'
        self.assertEqual(converter._fromBIN(
            '11000000111001', True), expected_output)

        expected_output = '{Decimal: 12345; Hexadecimal: 3039}'
        self.assertEqual(converter._fromBIN(
            '11000000111001', False), expected_output)
        
    def test_empty_input(self):
        expected_output = False
        self.assertEqual(converter.is_dec(''), expected_output)
        self.assertEqual(converter.is_hex(''), expected_output)
        self.assertEqual(converter.is_bin(''), expected_output)
        
    def test_correct_input(self):
        expected_output = True
        self.assertEqual(converter.is_dec('1'), expected_output)
        self.assertEqual(converter.is_dec('-1'), expected_output)
        self.assertEqual(converter.is_dec('999'), expected_output)
        
        self.assertEqual(converter.is_hex('1'), expected_output)
        self.assertEqual(converter.is_hex('-1'), expected_output)
        self.assertEqual(converter.is_hex('999'), expected_output)
        self.assertEqual(converter.is_hex('0x999'), expected_output)
        self.assertEqual(converter.is_hex('-0x999'), expected_output)
        
        self.assertEqual(converter.is_bin('1'), expected_output)
        self.assertEqual(converter.is_bin('-1'), expected_output)
        self.assertEqual(converter.is_bin('0101'), expected_output)
        self.assertEqual(converter.is_bin('0b0101'), expected_output)
        self.assertEqual(converter.is_bin('-0b0101'), expected_output)
        
    def test_wrong_input(self):
        expected_output = False
        self.assertEqual(converter.is_dec('0x1'), expected_output)
        self.assertEqual(converter.is_dec('1.5'), expected_output)
        
        self.assertEqual(converter.is_hex('FG'), expected_output)
        self.assertEqual(converter.is_hex('0x-999'), expected_output)
        
        self.assertEqual(converter.is_bin('2'), expected_output)
        self.assertEqual(converter.is_bin('0x1'), expected_output)

# python -m unittest discover -s tests -p test*.py



from glob import iglob
from os.path import isfile, realpath, isdir
from re import match

from cat_win.const.ArgConstants import *


FILE_ENCODING: str = 'utf-8'
FILE_SEARCH = []
FILE_MATCH = []
FILE_TRUNCATE = [None, None, None]


def __addArgument__(args: list, unknown_args: list, known_files: list, unknown_files: list, param: str) -> bool:
    """
    sorts an argument to either list option, by appending to it.
    
    Parameters:
    args (list):
        all known parameters
    unknown_args (list):
        all unknown parameters
    known_files (list):
        all known files
    unknown_files (list)
        all unknown files
    param (str):
        the current parameter
        
    Returns:
    (bool):
        True if -E has been called, meaning every following parameter
        should simply be printed to stdout.
    """
    # 'enc' + ('=' or ':') + FILE_ENCODING
    if match(r"\Aenc[\=\:].+\Z", param):
        global FILE_ENCODING
        FILE_ENCODING = param[4:]
        return False
    # 'match' + ('=' or ':') + FILE_MATCH
    elif match(r"\Amatch[\=\:].+\Z", param):
        global FILE_MATCH
        FILE_MATCH.append(fr'{param[6:]}')
        return False
    # 'find' + ('=' or ':') + FILE_SEARCH
    elif match(r"\Afind[\=\:].+\Z", param):
        global FILE_SEARCH
        FILE_SEARCH.append(param[5:])
        return False
    # 'trunc' + ('=' or ':') + FILE_TRUNCATE[0] + ':' + FILE_TRUNCATE[1] + ':' + FILE_TRUNCATE[2]
    elif match(r"\Atrunc[\=\:][0-9()+\-*\/]*\:[0-9()+\-*\/]*\:?[0-9()+\-*\/]*\Z", param):
        paramSplit = param[6:].split(':')
        global FILE_TRUNCATE
        FILE_TRUNCATE[0] = None if paramSplit[0] == '' else (
            0 if paramSplit[0] == '0' else int(eval(paramSplit[0]))-1)
        FILE_TRUNCATE[1] = None if paramSplit[1] == '' else int(eval(paramSplit[1]))
        if len(paramSplit) == 3:
            FILE_TRUNCATE[2] = None if paramSplit[2] == '' else int(eval(paramSplit[2]))
        return False
    # '[' + ARGS_CUT + ']'
    elif match(r"\A\[[0-9()+\-*\/]*\:[0-9()+\-*\/]*\:?[0-9()+\-*\/]*\]\Z", param):
        args.append((ARGS_CUT, param))
        return False
    # '[' + ARGS_REPLACE + ']'
    elif match(r"\A\[.+\,.+\]\Z", param):
        args.append((ARGS_REPLACE, param))
        return False

    # default parameters
    for x in ALL_ARGS:
        if x.shortForm == param or x.longForm == param:
            args.append((x.id, param))
            if x.id == ARGS_ECHO:
                return True
            return False

    possible_path = realpath(param)
    if match(r".*\*+.*", param):
        for filename in iglob(param, recursive=True):
            if isfile(filename):
                known_files.append(realpath(filename))
    elif isdir(possible_path):
        for filename in iglob(possible_path + '**/**', recursive=True):
            if isfile(filename):
                known_files.append(realpath(filename))
    elif isfile(possible_path):
        known_files.append(possible_path)
    elif len(param) > 2 and param[0] == '-' and param[1] != '-':
        for i in range(1, len(param)):
            if __addArgument__(args, unknown_args, known_files, unknown_files, '-' + param[i]):
                return True
    elif match(r"\A[^-]+.*\Z", param):
        unknown_files.append(realpath(param))
    else:
        unknown_args.append(param)
    return False


def getArguments(argv: list) -> tuple:
    """
    Read all args to either a valid parameter, an invalid parameter,
    a known file, an unknown file, or an echo parameter to print out.
    
    Parameters:
    argv (list):
        the entire sys.argv list
    
    Returns:
    (args, unknown_args, known_files, unknown_files, echo_args) (tuple):
        contains the paramater in a sorted manner
    """
    inputArgs = argv[1:]
    args = []
    unknown_args = []
    known_files = []
    unknown_files = []
    echo_args = []

    echoCall = False
    

    for arg in inputArgs:
        if echoCall:
            echo_args.append(arg)
            continue
        echoCall = __addArgument__(args, unknown_args, known_files, unknown_files, arg)

    return (args, unknown_args, known_files, unknown_files, echo_args)



from json import loads as loadJSON
from urllib.request import urlopen

from cat_win.const.ColorConstants import C_KW

from cat_win import __url__


# UNSAFE:
# UPDATE MAY INCLUDE FUNDAMENTAL CHANGES
# recognized by a higher version number in earlier position
# !.!._
# SAFE:
# recognized by a higher version number in the last position
# _._.!
STATUS_UP_TO_DATE = 0
STATUS_STABLE_RELEASE_AVAILABLE = 1
STATUS_UNSAFE_STABLE_RELEASE_AVAILABLE = -1
STATUS_PRE_RELEASE_AVAILABLE = 2
STATUS_UNSAFE_PRE_RELEASE_AVAILABLE = -2


def getLastestPackageVersion(package: str) -> str:
    """
    retrieve the official PythonPackageIndex information regarding
    a package.
    
    Parameters:
    package (str):
        the package name to check
        
    Returns:
    (str):
        a version representation
        on Error: a zero version '0.0.0
    """
    try:
        with urlopen(f"https://pypi.org/pypi/{package}/json", timeout=2) as _response:
            response = _response.read()
        return loadJSON(response)['info']['version']
    except:
        return '0.0.0'


def onlyNumeric(s: str) -> int:
    """
    strips every non-numeric character of a string.
    
    Parameters:
    s (str):
        the string to filter.
        
    Returns:
    (int):
        the resulting number of the string containing
        only numeric values
    """
    return int('0' + ''.join(filter(str.isdigit, s)))


def genVersionTuples(v: str, w: str) -> tuple:
    """
    create comparable version tuples.
    
    Parameters:
    v (str):
        a version representation like '1.0.33.0'
    w (str):
        a version representation like '1.1.0a'
    
    Returns:
    (tuple(tuple, tuple)):
        the version tuples of both inputs like
        (('01', '00', '33', '00'), ('01', '01', '0a', '00'))
    """
    vSplit, wSplit = v.split('.'), w.split('.')
    maxSplitLen = max(map(len, vSplit + wSplit))
    vList = [s.zfill(maxSplitLen) for s in vSplit]
    wList = [s.zfill(maxSplitLen) for s in wSplit]
    maxLength = max(len(vList), len(wList))
    vList += [''.zfill(maxSplitLen)] * (maxLength - len(vList))
    wList += [''.zfill(maxSplitLen)] * (maxLength - len(wList))
    return (tuple(vList), tuple(wList))

def newVersionAvailable(currentVersion: str, latestVersion: str) -> int:
    """
    Checks whether or not a new version is available.
    
    Parameters:
    currentVersion (str):
        a version representation as string
    latestVersion (str):
        a version representation as string
    
    Returns:
    (int):
        a global status code describing the situation
    """
    if currentVersion.startswith('v'):
        currentVersion = currentVersion[1:]
    if latestVersion.startswith('v'):
        latestVersion = latestVersion[1:]
    status = STATUS_UP_TO_DATE
    current, latest = genVersionTuples(currentVersion, latestVersion)
    i = 0
    for c, l in zip(current, latest):
        i += 1
        cNum, lNum = onlyNumeric(c), onlyNumeric(l)
        if cNum > lNum:
            break
        if cNum < lNum:
            status = STATUS_STABLE_RELEASE_AVAILABLE
            if not l.isdigit():
                status = STATUS_PRE_RELEASE_AVAILABLE
            break
        if c < l:
            status = STATUS_PRE_RELEASE_AVAILABLE
            break
    if i < len(current):
        status *= -1
    return status


def printUpdateInformation(package: str, currentVersion: str, color_dic: dict) -> None:
    """
    prints update information if there are any.
    
    Parameters:
    package (str):
        the package name to check
    currentVersion (str):
        a version representation as string of the current version
    color_dic (dict):
        a dictionary translating the color-keywords to ANSI-Colorcodes
    """
    latestVersion = getLastestPackageVersion(package)
    status = newVersionAvailable(currentVersion, latestVersion)
    if status == STATUS_UP_TO_DATE:
        return
    message = f""
    warning = f""
    info    = f""
    if abs(status) == STATUS_STABLE_RELEASE_AVAILABLE:
        message += f"{color_dic[C_KW.MESSAGE_IMPORTANT]}"
        message += f"A new stable release of {package} is available: v{latestVersion}"
        message += f"{color_dic[C_KW.RESET_ALL]}\n{color_dic[C_KW.MESSAGE_IMPORTANT]}"
        message += f"To update, run:"
        message += f"{color_dic[C_KW.RESET_ALL]}\n{color_dic[C_KW.MESSAGE_IMPORTANT]}"
        message += f"python -m pip install --upgrade {package}"
    elif abs(status) == STATUS_PRE_RELEASE_AVAILABLE:
        message += f"{color_dic[C_KW.MESSAGE_INFORMATION]}"
        message += f"A new pre-release of {package} is available: v{latestVersion}"
    message += f"{color_dic[C_KW.RESET_ALL]}"
    if status < STATUS_UP_TO_DATE:
        warning += f"{color_dic[C_KW.MESSAGE_WARNING]}"
        warning += f"Warning: Due to the drastic version increase, backwards compatibility is no longer guaranteed!"
        warning += f"{color_dic[C_KW.RESET_ALL]}\n{color_dic[C_KW.MESSAGE_WARNING]}"
        warning += f"You may experience fundamental differences."
        warning += f"{color_dic[C_KW.RESET_ALL]}"
    info += f"{color_dic[C_KW.MESSAGE_INFORMATION]}Take a look at the changelog here:"
    info += f"{color_dic[C_KW.RESET_ALL]}\n{color_dic[C_KW.MESSAGE_INFORMATION]}"
    info += f"{__url__}/blob/main/CHANGELOG.md{color_dic[C_KW.RESET_ALL]}"
    print(message)
    print(warning)
    print(info)




def getRawViewLinesGen(file: str = '', mode: str = 'X', colors = None, file_encoding: str = 'utf-8'):
    """
    return the raw byte representation of a file in hexadecimal or binary
    line by line
    
    Parameters:
    file (str):
        the file to use
    mode (str):
        either 'x', 'X' for hexadecimal (lower- or upper case letters),
        or 'b' for binary
    colors (list):
        a list of two elements. Index 0 holds the color C_KW.RAWVIEWER,
        Index 1 holds the color C_KW.RESET_ALL
    file_encoding (str):
        the encoding used (possibly for stdout)
        
    Yields:
    currentLine (str):
        a string representation that, when put together, forms the hexviewer
        output containing the header and line information aswell
        as the bytes themselves
    """
    if colors == None or len(colors) < 2:
        colors = ['', '']
    
    CRLF = {10: '␤', 13: '␍'}

    try:
        if len(CRLF[10].encode(file_encoding)) != 3:
            raise UnicodeEncodeError('', '', -1, -1, '')
    except UnicodeEncodeError:
        CRLF = {10: '·', 13: '·'}
    
    def getDisplayChar(byte: int) -> str:
        # 32 - 126 => ' ' - '~' (ASCII)
        if 32 <= byte <= 126:
            return chr(byte)
        elif byte in [10, 13]:
            return CRLF[byte]
        return '·'
    
    rawFile = open(file, 'rb')
    rawFileContent = rawFile.read()
    rawFileContentLength = len(rawFileContent)
    rawFile.close()
    
    reprLength = 2 * (mode.upper() == 'X') + 8 * (mode == 'b')
     
    currentLine = f"{colors[0]}Address  "
    for i in range(16):
        currentLine += f"{i:0{2}X} " + '      ' * (mode == 'b')
    currentLine += f"# Decoded Text                   {colors[1]}"
    yield currentLine
    
    currentLine = f"{colors[0]}{0:0{8}X}{colors[1]} "
    line = []
    for i, byte in enumerate(rawFileContent, start=1):
        line.append(byte)
        if not (i % 16):
            currentLine +=  ' '.join([f"{b:0{reprLength}{mode}}" for b in line]) + \
                            f" {colors[0]}#{colors[1]} " + \
                            ' '.join(map(getDisplayChar, line))
            yield currentLine
            if i < rawFileContentLength:
                currentLine = f"{colors[0]}{i:0{8}X}{colors[1]} "
            line = []
    if line:
        currentLine +=  ' '.join([f"{b:0{reprLength}{mode}}" for b in line]) + ' ' + \
                        ' ' * ((reprLength + 1) * (16-len(line)) - 1) + \
                        f" {colors[0]}#{colors[1]} " + \
                        ' '.join(map(getDisplayChar, line))
        yield currentLine



from base64 import b64encode, b64decode


def _encodeBase64(content: str, encoding: str) -> bytes:
    """
    Encode a string to base64.
    
    Parameters:
    content (str):
        the string to encode
    encoding (str):
        the encoding the string is using
        
    Returns:
    encoded_content (bytes):
        the base64 encoded string
    """
    # encode the string to bytes and encode with base64
    contentBytes = content.encode(encoding=encoding)
    encoded_content = b64encode(contentBytes)
    
    # return as a single line
    return encoded_content


def encodeBase64(content: list, encoding: str = 'utf-8') -> list:
    """
    Encode the file content to base64 by calling _encodeBase64()
    
    Parameters:
    content (list):
        the file content represented as [(prefix, line), ...]
    encoding (str):
        the file encoding used
        
    Returns:
    [('', encoded_content)] (list):
        the encoded base64 content as a single line without any prefix
    """
    # concatenate all lines and join them with line breaks
    contentLines = list(map(lambda x: x[1], content))
    contentLine = '\n'.join(contentLines)
    
    encoded_content = _encodeBase64(contentLine, encoding)
    encoded_content = encoded_content.decode(encoding='ascii')
    # return as a list containing a single line
    return [('', encoded_content)]


def _decodeBase64(content: str) -> bytes:
    """
    Decode a string from base64.
    
    Parameters:
    content (str):
        the string to decode
        
    Returns:
    decoded_content (bytes):
        the base64 decoded string
    """
    # encode the string to bytes and decode with base64
    base64_bytes = content.encode(encoding='ascii')
    decoded_content = b64decode(base64_bytes)
    
    # return as a single line
    return decoded_content


def decodeBase64(content: list, encoding: str = 'utf-8') -> list:
    """
    decode the file content from base64 by calling _decodeBase64()
    
    Parameters:
    content (list):
        the file content represented as [(prefix, line), ...]
    encoding (str):
        the file encoding used
        
    Returns:
    [('', line), ...] (list):
        the decoded base64 content line by line without any prefix
    """
    # concatenate all lines and join them
    contentLines = list(map(lambda x: x[1], content))
    contentLine = ''.join(contentLines)
    
    decoded_content = _decodeBase64(contentLine)
    decoded_content = decoded_content.decode(encoding=encoding)
    
    # return as content list, split at line breaks
    decoded_content = decoded_content.split('\n')
    return [('', line) for line in decoded_content]



from datetime import datetime
from math import log, pow, floor
from os import stat
from platform import system
from stat import (
    FILE_ATTRIBUTE_ARCHIVE as A,
    FILE_ATTRIBUTE_SYSTEM as S,
    FILE_ATTRIBUTE_HIDDEN as H,
    FILE_ATTRIBUTE_READONLY as R,
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED as I,
    FILE_ATTRIBUTE_COMPRESSED as C,
    FILE_ATTRIBUTE_ENCRYPTED as E
)


def _convert_size(size_bytes: int) -> str:
    """
    convert a size value to a more compact representation
    
    Parameters_
    size_bytes (int):
        a size value in bytes
        
    Returns:
    (str):
        a string representation with a size value suffix
    """
    if size_bytes == 0:
        return '0 B'
    size_name = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    i = int(floor(log(size_bytes, 1024)))
    p = pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i] if i < len(size_name) else '?'}"


def read_attribs(file: str) -> list:
    """
    check which attributes a file has set.
    
    Parameters:
    file (str):
        a string representation of a file (-path)
        
    Returns:
    (list):
        a list of lists containing attributes and a
        boolean value describing if it is set
        [[ATTRIBUTE, True/False], ...]
    """
    attrs = stat(file, follow_symlinks=False).st_file_attributes

    return (
        [['Archive', bool(attrs & A)],
         ['System', bool(attrs & S)],
         ['Hidden', bool(attrs & H)],
         ['Readonly', bool(attrs & R)],
         # Because this attribute is true when the file is _not_ indexed
         ['Indexed', not bool(attrs & I)],
         ['Compressed', bool(attrs & C)],
         ['Encrypted', bool(attrs & E)]]
    )


def getFileSize(file: str) -> int:
    """
    calculate the size of a file
    
    Parameters:
    file (str):
        a string representation of a file (-path)
        
    Returns:
    (int):
        the size in bytes or 0 if an (OS-)error occurs
    """
    try:
        return stat(file).st_size
    except OSError:
        return 0


def getFileMetaData(file: str, on_windows_os: bool, colors = None) -> str:
    """
    calculate file metadata information.
    
    Parameters:
    file (str):
        a string representation of a file (-path)
    on_windows_os (bool):
        indicates if the user is on windows os using
        platform.system() == 'Windows'
    colors (list):
        a list containing the ANSI-Colorcodes to display
        the attributes like [RESET_ALL, ATTRIB, +ATTRIB, -ATTRIB]
    
    Returns:
    metaData (str):
        representation containing file size, creation/modified/accessed time.
        on windows: also contains file attribute information
    """
    if colors == None or len(colors) < 4:
        colors = ['', '', '', '']
    try:
        stats = stat(file)
        
        metaData = colors[1] + file + colors[0] + '\n'
        
        metaData += f"{colors[1]}{'Size:' : <16}{_convert_size(stats.st_size)}{colors[0]}\n"
        metaData += f"{colors[1]}{'ATime:': <16}{datetime.fromtimestamp(stats.st_atime)}{colors[0]}\n"
        metaData += f"{colors[1]}{'MTime:': <16}{datetime.fromtimestamp(stats.st_mtime)}{colors[0]}\n"
        metaData += f"{colors[1]}{'CTime:': <16}{datetime.fromtimestamp(stats.st_ctime)}{colors[0]}\n"
        
        if not on_windows_os:
            metaData += '\n'
            return metaData
        
        attribs = read_attribs(file)
        metaData += colors[2]
        metaData += '+' + ", ".join([x for x, y in attribs if y])
        metaData += colors[0] + '\n'
        metaData += colors[3]
        metaData += '-' + ", ".join([x for x, y in attribs if not y])
        metaData += colors[0] + '\n'
        return metaData
    except OSError:
        return ''



from unittest import TestCase

import cat_win.web.UpdateChecker as UpdateChecker
# import sys
# sys.path.append('../cat_win')


class TestUpdateChecker(TestCase):
    def test_onlyNumeric(self):
        self.assertEqual(UpdateChecker.onlyNumeric('1nh589h15io125b085218'), 158915125085218)
        self.assertEqual(UpdateChecker.onlyNumeric('1nh'), 1)
        self.assertEqual(UpdateChecker.onlyNumeric('123'), 123)
        self.assertEqual(UpdateChecker.onlyNumeric('abc'), 0)
    
    def test_genVersionTuples(self):
        self.assertEqual(UpdateChecker.genVersionTuples('1.0.33.0', '1.1.0a'),
                         (('01', '00', '33', '00'), ('01', '01', '0a', '00')))
        self.assertEqual(UpdateChecker.genVersionTuples('1.2.3', '1.2.3.4.5'),
                         (('1', '2', '3', '0', '0'), ('1', '2', '3', '4', '5')))
        self.assertEqual(UpdateChecker.genVersionTuples('1.12345', '1.2'),
                         (('00001', '12345'), ('00001', '00002')))
    
    def test_version_comparison_up_to_date(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.0.9'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.0.09'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.10.9', 'v1.10.9'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.7a.9', '1.7a.9'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('09.0.8b', '9.0.08b'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('v7b.4', '07b.04.00'), 0)
        self.assertEqual(UpdateChecker.newVersionAvailable('9', '9.0.0'), 0)
        
    def test_version_comparison_stable_release(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.0.10'), 1)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.15', 'v1.0.016'), 1)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.12', '1.000.020'), 1)
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.0.0', 'v01.00.01'), 1)
        self.assertEqual(UpdateChecker.newVersionAvailable('2.1.5a', '2.1.6'), 1)
        self.assertEqual(UpdateChecker.newVersionAvailable('v2.1b.5a', '2.1b.051'), 1)

    def test_version_comparison_pre_release(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.0.9', '1.0.9a'), 2)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.0.09b'), 2)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', 'v1.0.10b'), 2)
        self.assertEqual(UpdateChecker.newVersionAvailable('2.1.5a', '2.1.5b'), 2)
    
    def test_version_comparison_stable_release_unsafe(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.01.10'), -1)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.15', '1.1.9'), -1)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', 'v2.0.9'), -1)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.12', '1.001.0'), -1)
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.0.0', 'v01.001.01'), -1)
        self.assertEqual(UpdateChecker.newVersionAvailable('2.1.5a', '2.2.5'), -1)

    def test_version_comparison_pre_release_unsafe(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.0.9', '1a.0.9a'), -2)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '1.0a.9'), -2)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', 'v1a.0.9'), -2)
        self.assertEqual(UpdateChecker.newVersionAvailable('1.0.9', '2b.0.10b'), -2)
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.0.9', '1.0b.10'), -2)
        self.assertEqual(UpdateChecker.newVersionAvailable('2.1.5a', '2.1a.5'), -2)
        
    def test_version_availability_current_is_newest(self):
        self.assertEqual(UpdateChecker.newVersionAvailable('v1.1.9', '1.0.9a'), 0)

# python -m unittest discover -s tests -p test*.py



from sys import stdin
import os


def writeTemp(content: str, tmp_file: str, file_encoding: str) -> str:
    """
    Writes content into a generated temp-file.
    
    Parameters:
    content (str):
        the content to write in a file
    tmp_file (str):
        a string representation of a file (-path)
    file_encoding (str):
        an encoding the open the file with
    
    Returns:
    tmp_file (str):
        the path to the temporary file written
    """
    with open(tmp_file, 'wb') as f:
        f.write(content.encode(file_encoding))
    return tmp_file


def getStdInContent(oneLine: bool = False):
    """
    read the stdin.
    
    Parameters:
    oneLine (bool):
        determines if only the first stdin line should be read
        
    Yields:
    line (str):
        the input (line by line) delivered by stdin
        until the first EOF (Chr(26)) character
    """
    if oneLine:
        first_line = stdin.readline().rstrip('\n')
        if first_line[-1:] == chr(26):
            first_line = first_line[:-1]
        yield first_line
        return
    for line in stdin:
        if line[-2:-1] == chr(26):
            yield line[:-2]
            break
        yield line


def path_parts(path: str) -> list:
    """
    split a path recursively into its parts.
    
    Parameters:
    path (str):
        a file/dir path
        
    Returns:
    (list):
        contains each drive/directory/file in the path seperated
        "C:/a/b/c/d.txt" -> ['C:/', 'a', 'b', 'c', 'd.txt']
    """
    p, f = os.path.split(path)
    return path_parts(p) + [f] if f and p else [p] if p else []


def createFile(file: str, content: str, file_encoding: str) -> bool:
    """
    create the directory path to a file, and the file itself.
    on error: cleanup all created subdirectories
    
    Parameters:
    file (str):
        a string representation of a file (-path)
    content (str):
        the content to write into the files
    file_encoding (str):
        the encoding to open the files with
    
    Returns:
    (bool):
        True if the operation was successful.
    """
    file_dir = os.path.dirname(file)
    splitted_path = path_parts(file_dir)
    subpaths = [os.path.join(*splitted_path[:i]) for i in range(2, len(splitted_path)+1)]
    unknown_subpaths = [s for s in subpaths[::-1] if not os.path.exists(s)]
    try:
        os.makedirs(file_dir, exist_ok=True)
    except OSError:
        print(f"Error: The path '{file_dir}' could not be created.")
        # cleanup (delete the folders that have been created)
        for subpath in unknown_subpaths:
            try:
                os.rmdir(subpath)
            except OSError:
                continue
        return False
    try:
        with open(file, 'wb') as f:
            f.write(content.encode(file_encoding))
    except OSError:
        print(f"Error: The file '{file}' could not be written.")
        # cleanup (delete the folders that have been created)
        for subpath in unknown_subpaths:
            try:
                os.rmdir(subpath)
            except OSError:
                continue
        return False
    return True


def writeFiles(file_list: list, content: str, file_encoding: str) -> list:
    """
    write to multiple files. ask if an empty file should be created
    when there is nothing to write.
    try to create the path to the files if it does not yet exist.
    delete the created path again (cleanup) if the file still could
    not be written.
    
    Parameters:
    file_list (list):
        all files that should be written
    content (str):
        the content to write into the files
    file_encoding (str):
        the encoding to open the files with
    
    Returns:
    (list):
        containing all files, that could succesfully be written.
    """
    if len(file_list) == 0:
        return file_list
    
    if content == '':
        abort_command = '' 
        try:
            print('You are about to create an empty file. Do you want to continue?')
            enterChar = '⏎'

            try:
                if len(enterChar.encode(file_encoding)) != 3:
                    raise UnicodeEncodeError('', '', -1, -1, '')
            except UnicodeEncodeError:
                enterChar = 'ENTER'
            print(f"[Y/{enterChar}] Yes, Continue       [N] No, Abort :", end='')
            abort_command = input()
        except EOFError:
            pass
        except UnicodeDecodeError:
            print(f"Input is not recognized in the given encoding: {file_encoding}")
            abort_command = 'n'
        finally:
            if abort_command and abort_command.upper() != 'Y':
                print('Aborting...')
                file_list.clear()
    
    success_file_list = []
    
    for file in file_list:
        try:
            with open(file, 'wb') as f:
                f.write(content.encode(file_encoding))
            success_file_list.append(file)
        except FileNotFoundError: # the os.pardir path to the file does not exist
            if createFile(file, content, file_encoding):
                success_file_list.append(file)
        except OSError:
            print(f"Error: The file '{file}' could not be written.")
            
    return success_file_list


def readWriteFilesFromStdIn(file_list: list, file_encoding: str, on_windows_os: bool, oneLine: bool = False) -> list:
    """
    Write stdin input to multiple files.
    
    Parameters:
    file_list (list):
        all files that should be written
    file_encoding (str):
        the encoding to use for writing the files
    on_windows_os (bool):
        indicates if the user is on windows os using
        platform.system() == 'Windows'
    oneLine (bool):
        determines if only the first stdin line should be read
        
    Returns:
    (list):
        containing all files, that could succesfully be written.
    """
    if len(file_list) == 0:
        return file_list

    print('The given FILE(s)', end='')
    print('', *file_list, sep='\n\t')
    EOFControlChar = 'Z' if on_windows_os else 'D'
    print(f"do/does not exist. Write the FILE(s) and finish with the ^{EOFControlChar}-suffix (Ctrl + {EOFControlChar}):")

    input = ''.join(getStdInContent(oneLine))

    return writeFiles(file_list, input, file_encoding)




class File:
    def __init__(self, path: str, displayName: str) -> None:
        self.path = path
        self.displayname = displayName
        
        self.containsQueried = False
        
    def setContainsQueried(self, containsQueried) -> None:
        self.containsQueried = self.containsQueried or containsQueried


from unittest import TestCase

from cat_win.util.StdInHelper import path_parts


class TestStdInHelper(TestCase):
    def test_path_parts(self):
        expected_output_win = ['C:/', 'a', 'b', 'c', 'd.txt']
        expected_output_unix_mac = ['a', 'b', 'c', 'd.txt']
        
        self.assertIn(path_parts('C:/a/b/c/d.txt'), [expected_output_win, expected_output_unix_mac])


from unittest import TestCase

from cat_win.util.StringFinder import StringFinder
# import sys
# sys.path.append('../cat_win')


class TestStringFinder(TestCase):
    def test_findliterals_True(self):
        stringFinder = StringFinder([], [])
        x = [x for x in stringFinder._findliterals('test', 'abctestdef')]
        self.assertEqual(x, [[3, 7]])
        
        x = [x for x in stringFinder._findliterals('test', 'testabctestdetestf')]
        self.assertEqual(x, [[0, 4], [7, 11], [13, 17]])
        
        x = [x for x in stringFinder._findliterals('', '')]
        self.assertEqual(x, [[0, 0]])
        
        x = [x for x in stringFinder._findliterals('', 'x')]
        self.assertEqual(x, [[0, 0], [1, 1]])
        
    def test_findLiterals_False(self):
        stringFinder = StringFinder([], [])
        x = [x for x in stringFinder._findliterals('a', '')]
        self.assertEqual(x, [])
        
        x = [x for x in stringFinder._findliterals('test', 'tsetabctesdeestf')]
        self.assertEqual(x, [])
        
    def test_findRegex_True(self):
        stringFinder = StringFinder([], [])
        x = [x for x in stringFinder._findregex(r"[0-9]{2}", '123')]
        self.assertEqual(x, [[0, 2]])
        
        x = [x for x in stringFinder._findregex(r"[0-9]{2}", '1234')]
        self.assertEqual(x, [[0, 2], [2, 4]])
        
        x = [x for x in stringFinder._findregex(r"[A-Z]{1}[a-z]*\s?.*\.+\s", 'Silas A. Kraume')]
        self.assertEqual(x, [[0, 9]])
        
        x = [x for x in stringFinder._findregex(r"[A-Z]{1}[a-z]*\s?.*\.+\s", 'silas A. Kraume')]
        self.assertEqual(x, [[6, 9]])
        
    def test_findRegex_False(self):
        stringFinder = StringFinder([], [])
        x = [x for x in stringFinder._findregex(r"[A-Z]{1}[a-z]+\s?.*\.+\s", 'silas A. Kraume')]
        self.assertEqual(x, [])
        
    def test_optimizeIntervals(self):
        stringFinder = StringFinder([], [])
        x = stringFinder._optimizeIntervals([[1, 3], [3, 5]])
        self.assertEqual(x, [[1, 5]])
        
        x = stringFinder._optimizeIntervals([[1, 10], [5, 10]])
        self.assertEqual(x, [[1, 10]])
        
        x = stringFinder._optimizeIntervals([[1, 3], [2, 5], [6, 8]])
        self.assertEqual(x, [[1, 5], [6, 8]])
        
        x = stringFinder._optimizeIntervals([[1, 3], [-5, 0], [0, 2]])
        self.assertEqual(x, [[-5, 3]])
        
    def test_optimizeIntervals_empty(self):
        stringFinder = StringFinder([], [])
        x = stringFinder._optimizeIntervals([])
        self.assertEqual(x, [])
        
    def test_findKeywords(self):
        stringFinder = StringFinder(['Is', 'Test', 'Not'], [r"[0-9]\!"])
        line = 'ThisIsATest!1!'
        intervals, fKeyWords, mKeywords = stringFinder.findKeywords(line)
        
        self.assertCountEqual(intervals, [[14, 'reset_matched'], [12, 'matched_pattern'], [11, 'reset_found'], [7, 'found_keyword'], [6, 'reset_found'], [4, 'found_keyword']])
        self.assertEqual(fKeyWords, [('Is', [4, 6]), ('Test', [7, 11])])
        self.assertEqual(mKeywords, [(r"[0-9]\!", [12, 14])])


import string


class Converter():
    def is_dec(self, v: str) -> bool:
        """
        Parameters:
        v (str):
            the string to check
            
        Returns:
            True if v is a Decimal number.
            False if it is not.
        """
        if v[:1] == '-':
            v = v[1:]
        return v.isdecimal() and v != ""

    def is_hex(self, v: str) -> bool:
        """
        Parameters:
        v (str):
            the string to check
            
        Returns:
            True if v is a Hexadecimal number.
            False if it is not.
        """
        if v[:1] == '-':
            v = v[1:]
        hex_digits = set(string.hexdigits)
        if v[:2] == '0x':
            v = v[2:]
        return all(c in hex_digits for c in v) and v != ""

    def is_bin(self, v: str) -> bool:
        """
        Parameters:
        v (str):
            the string to check
            
        Returns:
            True if v is a Binary number.
            False if it is not.
        """
        if v[:1] == '-':
            v = v[1:]
        if v[:2] == '0b':
            v = v[2:]
        v_set = set(v)
        return (v_set == {'0', '1'} or v_set == {'0'} or v_set == {'1'}) and v != ""

    def __dec_to_hex__(self, value: int, leading: bool = False) -> str:
        return '{0:#x}'.format(value) if leading else '{0:x}'.format(value)

    def __dec_to_bin__(self, value: int, leading: bool = False) -> str:
        return '{0:#b}'.format(value) if leading else '{0:b}'.format(value)

    def _fromDEC(self, value: int, leading: bool = False) -> str:
        """
        returns a String representation of a Decimal Int including the corresponding
        Hexadecimal and Binary number.
        """
        return '{Hexadecimal: ' + self.__dec_to_hex__(value, leading) + '; Binary: ' + self.__dec_to_bin__(value, leading) + '}'

    def __hex_to_dec__(self, value: str) -> str:
        return str(int(value, 16))

    def __hex_to_bin__(self, value: str, leading: bool = False) -> str:
        return bin(int(value, 16)) if leading else bin(int(value, 16))[2:]

    def _fromHEX(self, value: str, leading: bool = False) -> str:
        """
        returns a String representation of a Hexadecimal String including the corresponding
        Decimal and Binary number.
        """
        return '{Decimal: ' + self.__hex_to_dec__(value) + '; Binary: ' + self.__hex_to_bin__(value, leading) + '}'

    def __bin_to_dec__(self, value: str) -> str:
        return str(int(value, 2))

    def __bin_to_hex__(self, value: str, leading: bool = False) -> str:
        return hex(int(value, 2)) if leading else hex(int(value, 2))[2:]

    def _fromBIN(self, value: str, leading: bool = False) -> str:
        """
        returns a String representation of a Binary String including the corresponding
        Decimal and Hexadecimal number.
        """
        return '{Decimal: ' + self.__bin_to_dec__(value) + '; Hexadecimal: ' + self.__bin_to_hex__(value, leading) + '}'



from unittest import TestCase

from cat_win.const.ArgConstants import *
# import sys
# sys.path.append('../cat_win')


class TestArgConstants(TestCase):
    def test_unique_parameters(self):
        parameters_shortForm = [x.shortForm for x in ALL_ARGS]
        parameters_longForm = [x.longForm for x in ALL_ARGS]
        parameters_help = [x.help for x in ALL_ARGS]
        parameters_id= [x.id for x in ALL_ARGS]
        self.assertEqual(len(set(parameters_shortForm)), len(parameters_shortForm))
        self.assertEqual(len(set(parameters_longForm)), len(parameters_longForm))
        self.assertEqual(len(set(parameters_help)), len(parameters_help))
        self.assertEqual(len(set(parameters_id)), len(parameters_id))
        
    def test_no_concats(self):
        parameters_shortForm = [x.shortForm for x in ALL_ARGS]
        parameters_longForm = [x.longForm for x in ALL_ARGS]
        parameters = parameters_shortForm + parameters_longForm
        for param in parameters:
            self.assertEqual(param[:1], '-')
            if not param.startswith('--'):
                self.assertEqual(len(param), 2)